import numpy as np
import tensorflow as tf

import ants

def brain_extraction(image,
                     modality="t1v0",
                     antsxnet_cache_directory=None,
                     verbose=False):

    """
    Perform brain extraction using U-net and ANTs-based training data.  "NoBrainer"
    is also possible where brain extraction uses U-net and FreeSurfer training data
    ported from the

    https://github.com/neuronets/nobrainer-models

    Arguments
    ---------
    image : ANTsImage
        input image (or list of images for multi-modal scenarios).

    modality : string
        Modality image type.  Options include:
            * "t1": T1-weighted MRI---ANTs-trained.  Update from "t1v0".
            * "t1v0":  T1-weighted MRI---ANTs-trained.
            * "t1nobrainer": T1-weighted MRI---FreeSurfer-trained: h/t Satra Ghosh and Jakub Kaczmarzyk.
            * "t1combined": Brian's combination of "t1" and "t1nobrainer".  One can also specify
                            "t1combined[X]" where X is the morphological radius.  X = 12 by default.
            * "flair": FLAIR MRI.
            * "t2": T2 MRI.
            * "bold": 3-D BOLD MRI.
            * "fa": Fractional anisotropy.
            * "t1t2infant": Combined T1-w/T2-w infant MRI h/t Martin Styner.
            * "t1infant": T1-w infant MRI h/t Martin Styner.
            * "t2infant": T2-w infant MRI h/t Martin Styner.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    ANTs probability brain mask image.

    Example
    -------
    >>> probability_brain_mask = brain_extraction(brain_image, modality="t1")
    """

    from ..architectures import create_unet_model_3d
    from ..utilities import get_pretrained_network
    from ..utilities import get_antsxnet_data
    from ..architectures import create_nobrainer_unet_model_3d
    from ..utilities import decode_unet

    classes = ("background", "brain")
    number_of_classification_labels = len(classes)

    channel_size = 1
    if isinstance(image, list):
        channel_size = len(image)

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    input_images = list()
    if channel_size == 1:
        input_images.append(image)
    else:
        input_images = image

    if input_images[0].dimension != 3:
        raise ValueError( "Image dimension must be 3." )

    if "t1combined" in modality:
        # Need to change with voxel resolution
        morphological_radius = 12
        if '[' in modality and ']' in modality:
            morphological_radius = int(modality.split("[")[1].split("]")[0])

        brain_extraction_t1 = brain_extraction(image, modality="t1",
          antsxnet_cache_directory=antsxnet_cache_directory, verbose=verbose)
        brain_mask = ants.iMath_get_largest_component(
          ants.threshold_image(brain_extraction_t1, 0.5, 10000))
        brain_mask = ants.morphology(brain_mask,"close",morphological_radius).iMath_fill_holes()

        brain_extraction_t1nobrainer = brain_extraction(image * ants.iMath_MD(brain_mask, radius=morphological_radius),
          modality = "t1nobrainer", antsxnet_cache_directory=antsxnet_cache_directory, verbose=verbose)
        brain_extraction_combined = ants.iMath_fill_holes(
          ants.iMath_get_largest_component(brain_extraction_t1nobrainer * brain_mask))

        brain_extraction_combined = brain_extraction_combined + ants.iMath_ME(brain_mask, morphological_radius) + brain_mask

        return(brain_extraction_combined)

    if modality != "t1nobrainer":

        #####################
        #
        # ANTs-based
        #
        #####################

        weights_file_name_prefix = None

        if modality == "t1v0":
            weights_file_name_prefix = "brainExtraction"
        elif modality == "t1":
            weights_file_name_prefix = "brainExtractionT1v1"
        elif modality == "t2":
            weights_file_name_prefix = "brainExtractionT2"
        elif modality == "flair":
            weights_file_name_prefix = "brainExtractionFLAIR"
        elif modality == "bold":
            weights_file_name_prefix = "brainExtractionBOLD"
        elif modality == "fa":
            weights_file_name_prefix = "brainExtractionFA"
        elif modality == "t1t2infant":
            weights_file_name_prefix = "brainExtractionInfantT1T2"
        elif modality == "t1infant":
            weights_file_name_prefix = "brainExtractionInfantT1"
        elif modality == "t2infant":
            weights_file_name_prefix = "brainExtractionInfantT2"
        else:
            raise ValueError("Unknown modality type.")

        if verbose == True:
            print("Brain extraction:  retrieving model weights.")

        weights_file_name = get_pretrained_network(weights_file_name_prefix,
          antsxnet_cache_directory=antsxnet_cache_directory)

        if verbose == True:
            print("Brain extraction:  retrieving template.")

        reorient_template_file_name_path = get_antsxnet_data("S_template3",
          antsxnet_cache_directory=antsxnet_cache_directory)
        reorient_template = ants.image_read(reorient_template_file_name_path)
        resampled_image_size = reorient_template.shape

        number_of_filters = (8, 16, 32, 64)
        mode = "classification"
        if modality == "t1":
            number_of_filters = (16, 32, 64, 128)
            number_of_classification_labels = 1
            mode = "sigmoid"

        unet_model = create_unet_model_3d((*resampled_image_size, channel_size),
            number_of_outputs=number_of_classification_labels, mode=mode,
            number_of_filters=number_of_filters, dropout_rate=0.0,
            convolution_kernel_size=3, deconvolution_kernel_size=2,
            weight_decay=1e-5)

        unet_model.load_weights(weights_file_name)

        if verbose == True:
            print("Brain extraction:  normalizing image to the template.")

        center_of_mass_template = ants.get_center_of_mass(reorient_template)
        center_of_mass_image = ants.get_center_of_mass(input_images[0])
        translation = np.asarray(center_of_mass_image) - np.asarray(center_of_mass_template)
        xfrm = ants.create_ants_transform(transform_type="Euler3DTransform",
            center=np.asarray(center_of_mass_template), translation=translation)

        batchX = np.zeros((1, *resampled_image_size, channel_size))

        for i in range(len(input_images)):
            warped_image = ants.apply_ants_transform_to_image(xfrm, input_images[i], reorient_template)
            warped_array = warped_image.numpy()
            batchX[0,:,:,:,i] = (warped_array - warped_array.mean()) / warped_array.std()

        if verbose == True:
            print("Brain extraction:  prediction and decoding.")

        predicted_data = unet_model.predict(batchX, verbose=verbose)
        probability_images_array = decode_unet(predicted_data, reorient_template)

        if verbose == True:
            print("Brain extraction:  renormalize probability mask to native space.")

        xfrm_inv = xfrm.invert()
        probability_image = xfrm_inv.apply_to_image(probability_images_array[0][number_of_classification_labels-1], input_images[0])

        return(probability_image)

    else:

        #####################
        #
        # NoBrainer
        #
        #####################

        if verbose == True:
            print("NoBrainer:  generating network.")

        model = create_nobrainer_unet_model_3d((None, None, None, 1))

        weights_file_name = get_pretrained_network("brainExtractionNoBrainer",
          antsxnet_cache_directory=antsxnet_cache_directory)
        model.load_weights(weights_file_name)

        if verbose == True:
            print("NoBrainer:  preprocessing (intensity truncation and resampling).")

        image_array = image.numpy()
        image_robust_range = np.quantile(image_array[np.where(image_array != 0)], (0.02, 0.98))
        threshold_value = 0.10 * (image_robust_range[1] - image_robust_range[0]) + image_robust_range[0]

        thresholded_mask = ants.threshold_image(image, -10000, threshold_value, 0, 1)
        thresholded_image = image * thresholded_mask

        image_resampled = ants.resample_image(thresholded_image, (256, 256, 256), use_voxels=True)
        image_array = np.expand_dims(image_resampled.numpy(), axis=0)
        image_array = np.expand_dims(image_array, axis=-1)

        if verbose == True:
            print("NoBrainer:  predicting mask.")

        brain_mask_array = np.squeeze(model.predict(image_array, verbose=verbose))
        brain_mask_resampled = ants.copy_image_info(image_resampled, ants.from_numpy(brain_mask_array))
        brain_mask_image = ants.resample_image(brain_mask_resampled, image.shape, use_voxels=True, interp_type=1)

        spacing = ants.get_spacing(image)
        spacing_product = spacing[0] * spacing[1] * spacing[2]
        minimum_brain_volume = round(649933.7/spacing_product)
        brain_mask_labeled = ants.label_clusters(brain_mask_image, minimum_brain_volume)

        return(brain_mask_labeled)





import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Dropout, BatchNormalization,
                                     Activation, Dense, Concatenate,
                                     Conv2D, Conv2DTranspose, GlobalAveragePooling2D,
                                     Conv3D, Conv3DTranspose, GlobalAveragePooling3D)

from tensorflow.keras import initializers
from tensorflow.keras import regularizers

def create_densenet_model_2d(input_image_size,
                             number_of_classification_labels=1000,
                             number_of_filters=16,
                             depth=7,
                             number_of_dense_blocks=1,
                             growth_rate=12,
                             dropout_rate=0.2,
                             weight_decay=1e-4,
                             mode='classification'
                            ):
    """
    2-D implementation of the Wide ResNet deep learning architecture.

    Creates a keras model of the DenseNet deep learning architecture for image
    recognition based on the paper

    G. Huang, Z. Liu, K. Weinberger, and L. van der Maaten. Densely Connected
    Convolutional Networks Networks

    available here:

            https://arxiv.org/abs/1608.06993

    This particular implementation was influenced by the following python
    implementation:

            https://github.com/tdeboissiere/DeepLearningImplementations/blob/master/DenseNet/densenet.py

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    number_of_classification_labels : integer
        Number of classification labels.

    number_of_filters : integer
        Number of filters.

    depth : integer
        Number of layers---must be equal to 3 * N + 4 where N is an integer (default = 7).

    number_of_dense_blocks : integer
        Number of dense blocks number of dense blocks to add to the end (default = 1).

    growth_rate :  integer
        Number of filters to add for each dense block layer (default = 12).

    dropout_rate : scalar
        Per drop out layer rate (default = 0.2).

    weight_decay : scalar
        Weight decay (default = 1e-4).

    mode : string
        'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_densenet_model_2d((128, 128, 1))
    >>> model.summary()
    """

    concatenation_axis = 0
    if K.image_data_format() == 'channels_last':
        concatenation_axis = -1

    def convolution_factory_2d(model, number_of_filters, kernel_size=(3, 3),
                               dropout_rate=0.0, weight_decay=1e-4):
        model = BatchNormalization(axis=concatenation_axis,
                                   gamma_regularizer=regularizers.l2(weight_decay),
                                   beta_regularizer=regularizers.l2(weight_decay))(model)
        model = Activation(activation='relu')(model)
        model = Conv2D(filters=number_of_filters,
                       kernel_size=kernel_size,
                       padding='same',
                       use_bias=False,
                       kernel_initializer=initializers.he_normal(),
                       kernel_regularizer=regularizers.l2(weight_decay))(model)
        if dropout_rate > 0.0:
            model = Dropout(rate=dropout_rate)(model)

        return(model)

    def transition_2d(model, number_of_filters, dropout_rate=0.0, weight_decay=1e-4):

        model = convolution_factory_2d(model, number_of_filters, kernel_size=(1, 1),
                                       dropout_rate=dropout_rate, weight_decay=weight_decay)
        model = AveragePooling2D(pool_size=(2, 2),
                                 strides=(2, 2))(model)
        return(model)

    def create_dense_blocks_2d(model, number_of_filters, depth, growth_rate,
                               dropout_rate=0.0, weight_decay=1e-4):

        dense_block_layers = [model]
        for i in range(depth):
            model = convolution_factory_2d(model, number_of_filters=growth_rate,
                                           kernel_size=(3, 3), dropout_rate=dropout_rate,
                                           weight_decay=weight_decay)
            dense_block_layers.append(model)
            model = Concatenate(axis=concatenation_axis)(dense_block_layers)
            number_of_filters += growth_rate

        return(model, number_of_filters)


    if ((depth - 4) % 3) != 0:
        raise ValueError('Depth must be equal to 3*N+4 where N is an integer.')

    number_of_layers = int((depth - 4) / 3)

    inputs = Input(shape = input_image_size)

    outputs = Conv2D(filters=number_of_filters,
                     kernel_size=(3, 3),
                     kernel_initializer='he_uniform',
                     padding='same',
                     use_bias=False,
                     kernel_regularizer=regularizers.l2(weight_decay))(inputs)

    # Add dense blocks

    nFilters = number_of_filters

    for i in range(number_of_dense_blocks - 1):
        outputs, nFilters = \
          create_dense_blocks_2d(outputs, number_of_filters=nFilters,
                                 depth=number_of_layers, growth_rate=growth_rate,
                                 dropout_rate=dropout_rate, weight_decay=weight_decay)
        outputs = transition_2d(outputs, number_of_filters=nFilters,
                                dropout_rate=dropout_rate, weight_decay=weight_decay)

    outputs, nFilters = \
      create_dense_blocks_2d(outputs, number_of_filters=nFilters,
                             depth=number_of_layers, growth_rate=growth_rate,
                             dropout_rate=dropout_rate, weight_decay=weight_decay)

    outputs = BatchNormalization(axis=concatenation_axis,
                                 gamma_regularizer=regularizers.l2(weight_decay),
                                 beta_regularizer=regularizers.l2(weight_decay))(outputs)


    outputs = Activation(activation='relu')(outputs)
    outputs = GlobalAveragePooling2D()(outputs)

    layer_activation = ''
    if mode == 'classification':
        layer_activation = 'softmax'
    elif mode == 'regression':
        layerActivation = 'linear'
    else:
        raise ValueError('mode must be either `classification` or `regression`.')

    outputs = Dense(units=number_of_classification_labels,
                    activation=layer_activation,
                    kernel_regularizer=regularizers.l2(weight_decay),
                    bias_regularizer=regularizers.l2(weight_decay))(outputs)

    densenet_model = Model(inputs=inputs, outputs=outputs)

    return(densenet_model)

def create_densenet_model_3d(input_image_size,
                             number_of_classification_labels=1000,
                             number_of_filters=16,
                             depth=7,
                             number_of_dense_blocks=1,
                             growth_rate=12,
                             dropout_rate=0.2,
                             weight_decay=1e-4,
                             mode='classification'
                            ):
    """
    2-D implementation of the Wide ResNet deep learning architecture.

    Creates a keras model of the DenseNet deep learning architecture for image
    recognition based on the paper

    G. Huang, Z. Liu, K. Weinberger, and L. van der Maaten. Densely Connected
    Convolutional Networks Networks

    available here:

            https://arxiv.org/abs/1608.06993

    This particular implementation was influenced by the following python
    implementation:

            https://github.com/tdeboissiere/DeepLearningImplementations/blob/master/DenseNet/densenet.py

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    number_of_classification_labels : integer
        Number of classification labels.

    number_of_filters : integer
        Number of filters.

    depth : integer
        Number of layers---must be equal to 3 * N + 4 where N is an integer (default = 7).

    number_of_dense_blocks : integer
        Number of dense blocks number of dense blocks to add to the end (default = 1).

    growth_rate :  integer
        Number of filters to add for each dense block layer (default = 12).

    dropout_rate : scalar
        Per drop out layer rate (default = 0.2).

    weight_decay : scalar
        Weight decay (default = 1e-4).

    mode :  string
        'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_densenet_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    concatenation_axis = 0
    if K.image_data_format() == 'channels_last':
        concatenation_axis = -1

    def convolution_factory_3d(model, number_of_filters, kernel_size=(3, 3, 3),
                               dropout_rate=0.0, weight_decay=1e-4):
        model = BatchNormalization(axis=concatenation_axis,
                                   gamma_regularizer=regularizers.l2(weight_decay),
                                   beta_regularizer=regularizers.l2(weight_decay))(model)
        model = Activation(activation='relu')(model)
        model = Conv3D(filters=number_of_filters,
                       kernel_size=kernel_size,
                       padding='same',
                       use_bias=False,
                       kernel_initializer=initializers.he_normal(),
                       kernel_regularizer=regularizers.l2(weight_decay))(model)
        if dropout_rate > 0.0:
            model = Dropout(rate=dropout_rate)(model)

        return(model)

    def transition_3d(model, number_of_filters, dropout_rate=0.0, weight_decay=1e-4):

        model = convolution_factory_3d(model, number_of_filters, kernel_size=(1, 1, 1),
                                       dropout_rate=dropout_rate, weight_decay=weight_decay)
        model = AveragePooling3D(pool_size=(2, 2, 2),
                                 strides=(2, 2, 2))(model)
        return(model)

    def create_dense_blocks_3d(model, number_of_filters, depth, growth_rate,
                               dropout_rate=0.0, weight_decay=1e-4):

        dense_block_layers = [model]
        for i in range(depth):
            model = convolution_factory_3d(model, number_of_filters=growth_rate,
                                           kernel_size=(3, 3, 3), dropout_rate=dropout_rate,
                                           weight_decay=weight_decay)
            dense_block_layers.append(model)
            model = Concatenate(axis=concatenation_axis)(dense_block_layers)
            number_of_filters += growth_rate

        return(model, number_of_filters)


    if ((depth - 4) % 3) != 0:
        raise ValueError('Depth must be equal to 3*N+4 where N is an integer.')

    number_of_layers = int((depth - 4) / 3)

    inputs = Input(shape = input_image_size)

    outputs = Conv3D(filters=number_of_filters,
                     kernel_size=(3, 3, 3),
                     kernel_initializer='he_uniform',
                     padding='same',
                     use_bias=False,
                     kernel_regularizer=regularizers.l2(weight_decay))(inputs)

    # Add dense blocks

    nFilters = number_of_filters

    for i in range(number_of_dense_blocks - 1):
        outputs, nFilters = \
          create_dense_blocks_3d(outputs, number_of_filters=nFilters,
                                 depth=number_of_layers, growth_rate=growth_rate,
                                 dropout_rate=dropout_rate, weight_decay=weight_decay)
        outputs = transition_3d(outputs, number_of_filters=nFilters,
                                dropout_rate=dropout_rate, weight_decay=weight_decay)

    outputs, nFilters = \
      create_dense_blocks_3d(outputs, number_of_filters=nFilters,
                             depth=number_of_layers, growth_rate=growth_rate,
                             dropout_rate=dropout_rate, weight_decay=weight_decay)

    outputs = BatchNormalization(axis=concatenation_axis,
                                 gamma_regularizer=regularizers.l2(weight_decay),
                                 beta_regularizer=regularizers.l2(weight_decay))(outputs)


    outputs = Activation(activation='relu')(outputs)
    outputs = GlobalAveragePooling3D()(outputs)

    layer_activation = ''
    if mode == 'classification':
        layer_activation = 'softmax'
    elif mode == 'regression':
        layerActivation = 'linear'
    else:
        raise ValueError('mode must be either `classification` or `regression`.')

    outputs = Dense(units=number_of_classification_labels,
                    activation=layer_activation,
                    kernel_regularizer=regularizers.l2(weight_decay),
                    bias_regularizer=regularizers.l2(weight_decay))(outputs)

    densenet_model = Model(inputs=inputs, outputs=outputs)

    return(densenet_model)






from sklearn.preprocessing import PolynomialFeatures
from sklearn.linear_model import LinearRegression

import numpy as np

import ants

def regression_match_image(source_image,
                           reference_image,
                           mask=None,
                           poly_order=1,
                           truncate=True):
    """
    Image intensity normalization using linear regression.

    Arguments
    ---------
    source_image : ANTsImage
        Image whose intensities are matched to the reference
        image.

    reference_image : ANTsImage
        Defines the reference intensity function.

    poly_order : integer
        Polynomial order of fit.  Default is 1 (linear fit).

    mask : ANTsImage
        Defines voxels for regression modeling.

    truncate : boolean
        Turns on/off the clipping of intensities.

    Returns
    -------
    ANTs image (i.e., source_image) matched to the (reference_image).

    Example
    -------
    >>> import ants
    >>> source_image = ants.image_read(ants.get_ants_data('r16'))
    >>> reference_image = ants.image_read(ants.get_ants_data('r64'))
    >>> matched_image = regression_match_image(source_image, reference_image)
    """

    if source_image.shape != reference_image.shape:
        raise ValueError("Images do not have the same dimension.")

    if mask is None:
        source_intensities = np.expand_dims((source_image.numpy()).flatten(), axis=1)
        reference_intensities = np.expand_dims((reference_image.numpy()).flatten(), axis=1)
    else:
        mask_intensities = (mask.numpy()).flatten()
        source_intensities = np.expand_dims(source_image.numpy().flatten()[np.where(mask_intensities != 0)], axis=1)
        reference_intensities = np.expand_dims(source_image.numpy().flatten()[np.where(mask_intensities != 0)], axis=1)

    poly_features = PolynomialFeatures(degree=poly_order)
    source_intensities_poly = poly_features.fit_transform(source_intensities)

    model = LinearRegression()
    model.fit(source_intensities_poly, reference_intensities)

    matched_source_intensities = model.predict(source_intensities_poly)

    if truncate == True:
        min_reference_value = reference_intensities.min()
        max_reference_value = reference_intensities.max()
        matched_source_intensities[matched_source_intensities < min_reference_value] = min_reference_value
        matched_source_intensities[matched_source_intensities > max_reference_value] = max_reference_value

    matched_source_image = ants.make_image(source_image.shape, matched_source_intensities)
    matched_source_image = ants.copy_image_info(source_image,  matched_source_image)

    return(matched_source_image)




from .create_autoencoder_model import create_autoencoder_model
from .create_convolutional_autoencoder_model import create_convolutional_autoencoder_model_2d, create_convolutional_autoencoder_model_3d

from .create_alexnet_model import create_alexnet_model_2d, create_alexnet_model_3d
from .create_densenet_model import create_densenet_model_2d, create_densenet_model_3d
from .create_resnet_model import create_resnet_model_2d, create_resnet_model_3d
from .create_simple_classification_with_spatial_transformer_network_model import create_simple_classification_with_spatial_transformer_network_model_2d, create_simple_classification_with_spatial_transformer_network_model_3d

from .create_deep_back_projection_network_model import create_deep_back_projection_network_model_2d, create_deep_back_projection_network_model_3d
from .create_deep_denoise_super_resolution_model import create_deep_denoise_super_resolution_model_2d, create_deep_denoise_super_resolution_model_3d
from .create_denoising_auto_encoder_super_resolution_model import create_denoising_auto_encoder_super_resolution_model_2d, create_denoising_auto_encoder_super_resolution_model_3d
from .create_expanded_super_resolution_model import create_expanded_super_resolution_model_2d, create_expanded_super_resolution_model_3d
from .create_image_super_resolution_model import create_image_super_resolution_model_2d, create_image_super_resolution_model_3d
from .create_resnet_super_resolution_model import create_resnet_super_resolution_model_2d, create_resnet_super_resolution_model_3d
from .create_vgg_model import create_vgg_model_2d, create_vgg_model_3d
from .create_wide_resnet_model import create_wide_resnet_model_2d, create_wide_resnet_model_3d

from .create_unet_model import create_unet_model_2d, create_unet_model_3d
from .create_resunet_model import create_resunet_model_2d, create_resunet_model_3d
from .create_denseunet_model import create_denseunet_model_2d, create_denseunet_model_3d
from .create_custom_unet_model import create_nobrainer_unet_model_3d
from .create_custom_unet_model import create_hippmapp3r_unet_model_3d
from .create_custom_unet_model import create_sysu_media_unet_model_2d
from .create_custom_unet_model import create_hypothalamus_unet_model_3d

from .create_custom_model import create_simple_fully_convolutional_network_model_3d

from .create_vanilla_gan_model import VanillaGanModel
from .create_deep_convolutional_gan_model import DeepConvolutionalGanModel
from .create_wasserstein_gan_model import WassersteinGanModel
from .create_improved_wasserstein_gan_model import ImprovedWassersteinGanModel
from .create_cycle_gan_model import CycleGanModel
from .create_super_resolution_gan_model import SuperResolutionGanModel





import tensorflow.keras.backend as K
import tensorflow as tf

from tensorflow.keras.layers import Layer, InputSpec

#################################################################
#
#  Resampling layers (to a fixed size)
#
#################################################################

class ResampleTensorLayer2D(Layer):

    """
    Tensor resampling layer (2D).

    Arguments
    ---------
    shape : tuple
        Specifies the output shape of the resampled tensor.

    interpolation_type : string
        One of 'nearest_neighbor', 'linear', or 'cubic'.

    Returns
    -------
    Keras layer
        A keras layer

    """

    def __init__(self, shape, interpolation_type='nearest_neighbor', name='', **kwargs):

        if len(shape) != 2:
            raise ValueError("shape must be of length 2 specifying the width and " +
                             "height of the resampled tensor.")
        self.shape = shape

        allowed_types = set(['nearest_neighbor', 'linear', 'cubic'])
        if not interpolation_type in allowed_types:
            raise ValueError("interpolation_type not one of the allowed types.")
        self.interpolation_type = interpolation_type

        self._name = name

        super(ResampleTensorLayer2D, self).__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        if len(input_shape) != 4:
            raise ValueError("Input tensor must be of rank 4.")
        return((input_shape[0], self.shape[0], self.shape[1], input_shape[3]))

    def call(self, x, mask=None):
        resampled_tensor = None
        if self.interpolation_type == 'nearest_neighbor':
            resampled_tensor = tf.image.resize(x, size=self.shape, method='nearest')
        elif self.interpolation_type == 'linear':
            resampled_tensor = tf.image.resize(x, size=self.shape, method='bilinear')
        elif self.interpolation_type == 'cubic':
            resampled_tensor = tf.image.resize(x, size=self.shape, method='bicubic')
        return(resampled_tensor)

    def get_config(self):
        config = {"shape": self.shape, "interpolation_type": self.interpolation_type}
        return dict(list(config.items()))

class ResampleTensorLayer3D(Layer):

    """
    Tensor resampling layer (3D).

    Arguments
    ---------
    shape : tuple
        Specifies the output shape of the resampled tensor.

    interpolation_type : string
        One of 'nearest_neighbor', 'linear', or 'cubic'.

    Returns
    -------
    Keras layer
        A keras layer

    """

    def __init__(self, shape, interpolation_type='nearest_neighbor', name='', **kwargs):

        if len(shape) != 3:
            raise ValueError("shape must be of length 3 specifying the width, " +
                             "height, and depth of the resampled tensor.")
        self.shape = shape

        allowed_types = set(['nearest_neighbor', 'linear', 'cubic'])
        if not interpolation_type in allowed_types:
            raise ValueError("interpolation_type not one of the allowed types.")
        self.interpolation_type = interpolation_type

        self._name = name

        super(ResampleTensorLayer3D, self).__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        if len(input_shape) != 5:
            raise ValueError("Input tensor must be of rank 5.")
        return((input_shape[0], self.shape[0], self.shape[1], self.shape[2], input_shape[4]))

    def call(self, x, mask=None):

        channel_size = x.get_shape()[-1]

        resampled_tensor = None

        # Do yz

        new_shape_squeeze_yz = (-1, tf.shape(x)[2], tf.shape(x)[3], channel_size)
        squeeze_tensor_yz = tf.reshape(x, new_shape_squeeze_yz)

        resampled_tensor_yz = None
        new_shape_yz = (self.shape[1], self.shape[2])
        if self.interpolation_type == 'nearest_neighbor':
            resampled_tensor_yz = tf.image.resize(squeeze_tensor_yz, size=new_shape_yz, method='nearest')
        elif self.interpolation_type == 'linear':
            resampled_tensor_yz = tf.image.resize(squeeze_tensor_yz, size=new_shape_yz, method='bilinear')
        elif self.interpolation_type == 'cubic':
            resampled_tensor_yz = tf.image.resize(squeeze_tensor_yz, size=new_shape_yz, method='bicubic')

        new_shape_yz = (-1, tf.shape(x)[1], self.shape[1], self.shape[2], channel_size)
        resume_tensor_yz = tf.reshape(resampled_tensor_yz, new_shape_yz)

        # Do x

        reoriented_tensor = tf.transpose(resume_tensor_yz, (0, 3, 2, 1, 4))

        new_shape_squeeze_x = (-1, self.shape[1], tf.shape(x)[1], channel_size)
        squeeze_tensor_x = tf.reshape(reoriented_tensor, new_shape_squeeze_x)

        resampled_tensor_x = None
        new_shape_x = (self.shape[1], self.shape[0])
        if self.interpolation_type == 'nearest_neighbor':
            resampled_tensor_x = tf.image.resize(squeeze_tensor_x, size=new_shape_x, method='nearest')
        elif self.interpolation_type == 'linear':
            resampled_tensor_x = tf.image.resize(squeeze_tensor_x, size=new_shape_x, method='bilinear')
        elif self.interpolation_type == 'cubic':
            resampled_tensor_x = tf.image.resize(squeeze_tensor_x, size=new_shape_x, method='bicubic')

        new_shape_x = (-1, self.shape[2], self.shape[1], self.shape[0], channel_size)
        resumeTensor_x = tf.reshape(resampled_tensor_x, new_shape_x)

        resampled_tensor = tf.transpose(resumeTensor_x, (0, 3, 2, 1, 4))

        return(resampled_tensor)

    def get_config(self):
        config = {"shape": self.shape, "interpolation_type": self.interpolation_type}
        return dict(list(config.items()))

#################################################################
#
#  Resampling layers (to a target tensor)
#
#################################################################

class ResampleTensorToTargetTensorLayer2D(Layer):

    """
    Tensor resampling layer (2D).

    Arguments
    ---------

    interpolation_type : string
        One of 'nearest_neighbor', 'linear', or 'cubic'.

    Returns
    -------
    Keras layer
        A keras layer

    """

    def __init__(self, interpolation_type='nearest_neighbor', name='', **kwargs):

        allowed_types = set(['nearest_neighbor', 'linear', 'cubic'])
        if not interpolation_type in allowed_types:
            raise ValueError("interpolation_type not one of the allowed types.")
        self.interpolation_type = interpolation_type

        self._name = name
        self.resampled_tensor = None

        super(ResampleTensorToTargetTensorLayer2D, self).__init__(**kwargs)

    def call(self, x, mask=None):

        source_tensor = x[0]
        target_tensor = x[1]

        new_shape = (tf.shape(target_tensor)[1], tf.shape(target_tensor)[2])

        if self.interpolation_type == 'nearest_neighbor':
            self.resampled_tensor = tf.image.resize(source_tensor, size=new_shape, method='nearest')
        elif self.interpolation_type == 'linear':
            self.resampled_tensor = tf.image.resize(source_tensor, size=new_shape, method='bilinear')
        elif self.interpolation_type == 'cubic':
            self.resampled_tensor = tf.image.resize(source_tensor, size=new_shape, method='bicubic')
        return(self.resampled_tensor)

    def compute_output_shape(self, input_shape):
        return(K.int_shape(self.resampled_tensor))

    def get_config(self):
        config = {"interpolation_type": self.interpolation_type}
        return dict(list(config.items()))

class ResampleTensorToTargetTensorLayer3D(Layer):

    """
    Tensor resampling layer (3D).

    Arguments
    ---------

    interpolation_type : string
        One of 'nearest_neighbor', 'linear', or 'cubic'.

    Returns
    -------
    Keras layer
        A keras layer

    """

    def __init__(self, interpolation_type='nearest_neighbor', name='', **kwargs):

        allowed_types = set(['nearest_neighbor', 'linear', 'cubic'])
        if not interpolation_type in allowed_types:
            raise ValueError("interpolation_type not one of the allowed types.")
        self.interpolation_type = interpolation_type

        self._name = name
        self.resampled_tensor = None

        super(ResampleTensorToTargetTensorLayer3D, self).__init__(**kwargs)

    def call(self, x, mask=None):

        source_tensor = x[0]
        target_tensor = x[1]

        channel_size = source_tensor.get_shape()[-1]

        # Do yz

        new_shape_squeeze_yz = (-1, tf.shape(source_tensor)[2], tf.shape(source_tensor)[3], channel_size)
        squeeze_tensor_yz = tf.reshape(source_tensor, new_shape_squeeze_yz)

        resampled_tensor_yz = None
        new_shape_yz = (tf.shape(target_tensor)[2], tf.shape(target_tensor)[3])
        if self.interpolation_type == 'nearest_neighbor':
            resampled_tensor_yz = tf.image.resize(squeeze_tensor_yz, size=new_shape_yz, method='nearest')
        elif self.interpolation_type == 'linear':
            resampled_tensor_yz = tf.image.resize(squeeze_tensor_yz, size=new_shape_yz, method='bilinear')
        elif self.interpolation_type == 'cubic':
            resampled_tensor_yz = tf.image.resize(squeeze_tensor_yz, size=new_shape_yz, method='bicubic')

        new_shape_yz = (-1, tf.shape(source_tensor)[1], tf.shape(target_tensor)[2], tf.shape(target_tensor)[3], channel_size)
        resume_tensor_yz = tf.reshape(resampled_tensor_yz, new_shape_yz)

        # Do x

        reoriented_tensor = tf.transpose(resume_tensor_yz, (0, 3, 2, 1, 4))

        new_shape_squeeze_x = (-1, tf.shape(target_tensor)[2], tf.shape(source_tensor)[1], channel_size)
        squeeze_tensor_x = tf.reshape(reoriented_tensor, new_shape_squeeze_x)

        resampled_tensor_x = None
        new_shape_x = (tf.shape(target_tensor)[2], tf.shape(target_tensor)[1])
        if self.interpolation_type == 'nearest_neighbor':
            resampled_tensor_x = tf.image.resize(squeeze_tensor_x, size=new_shape_x, method='nearest')
        elif self.interpolation_type == 'linear':
            resampled_tensor_x = tf.image.resize(squeeze_tensor_x, size=new_shape_x, method='bilinear')
        elif self.interpolation_type == 'cubic':
            resampled_tensor_x = tf.image.resize(squeeze_tensor_x, size=new_shape_x, method='bicubic')

        new_shape_x = (-1, tf.shape(target_tensor)[3], tf.shape(target_tensor)[2], tf.shape(target_tensor)[1], channel_size)
        resumeTensor_x = tf.reshape(resampled_tensor_x, new_shape_x)

        self.resampled_tensor = tf.transpose(resumeTensor_x, (0, 3, 2, 1, 4))

        return(self.resampled_tensor)

    def compute_output_shape(self, input_shape):
        return(K.int_shape(self.resampled_tensor))

    def get_config(self):
        config = {"interpolation_type": self.interpolation_type}
        return dict(list(config.items()))





import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Lambda, Concatenate, Flatten, Dense,
                                    Conv2D, Conv2DTranspose, MaxPooling2D,
                                    ZeroPadding2D,
                                    Conv3D, Conv3DTranspose, MaxPooling3D,
                                    ZeroPadding3D)

def create_alexnet_model_2d(input_image_size,
                            number_of_classification_labels=1000,
                            number_of_dense_units=4096,
                            dropout_rate=0.0,
                            mode='classification'):
    """
    2-D implementation of the AlexNet deep learning architecture.

    Creates a keras model of the AlexNet deep learning architecture for image
    recognition based on the paper

    A. Krizhevsky, and I. Sutskever, and G. Hinton. ImageNet Classification with Deep Convolutional Neural Networks.

    available here:

            http://papers.nips.cc/paper/4824-imagenet-classification-with-deep-convolutional-neural-networks.pdf

    This particular implementation was influenced by the following python
    implementation:

            https://github.com/duggalrahul/AlexNet-Experiments-Keras/
            https://github.com/lunardog/convnets-keras/

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The
        shape (or dimension) of that tensor is the image dimensions followed by
        the number of channels (e.g., red, green, and blue).  The batch size
        (i.e., number of training images) is not specified a priori.

    number_of_classification_labels : integer
        Number of segmentation labels.

    number_of_dense_units : integer
       Number of dense units.

    dropout_rate : scalar
       Optional regularization parameter between [0, 1]. Default = 0.0.

    mode : string
       'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_alexnet_model_2d((128, 128, 1))
    >>> model.summary()

    """

    def split_tensor_2d(axis=1, ratio_split=1, id_split=0):

        def f(X):
           Xdims = K.int_shape(X)
           div = Xdims[axis] // ratio_split
           axis_split = slice(id_split * div, (id_split + 1) * div)

           if axis == 0:
               output = X[axis_split,:, :, :]
           elif axis == 1:
               output = X[:,axis_split, :, :]
           elif axis == 2:
               output = X[:,:, axis_split, :]
           elif axis == 3:
               output = X[:,:, :, axis_split]
           else:
               raise ValueError("axis specified is out of range.")

           return(output)

        def g(input_shape):
            output_shape = list(input_shape)
            output_shape[axis] = output_shape[axis] // ratio_split
            return(tuple(output_shape))

        return(Lambda(f, output_shape=lambda input_shape:g(input_shape)))

    def cross_channel_normalization_2d(alpha=1e-4, k=2, beta=0.75, n=5):

        def normalize_tensor_2d(X):

            X2 = K.square(X)

            half = n // 2

            extra_channels = K.spatial_2d_padding(
                K.permute_dimensions(X2, (1, 2, 3, 0)),
                padding=((0, 0), (half, half)))
            extra_channels = K.permute_dimensions(
                extra_channels, (3, 0, 1, 2))

            Xdims = K.int_shape(X)
            number_of_channels = int(Xdims[-1])

            scale = k
            for i in range(n):
                scale += alpha * extra_channels[:,:,:,i:(i + number_of_channels)]
            scale = scale ** beta

            return(X / scale)

        return(Lambda(normalize_tensor_2d, output_shape = lambda input_shape:input_shape))

    inputs = Input(shape = input_image_size)

    # Conv1
    outputs = Conv2D(filters = 96,
                     kernel_size=(11, 11),
                     strides=(4, 4),
                     activation='relu')(inputs)

    # Conv2
    outputs = MaxPooling2D(pool_size=(3, 3),
                           strides=(2, 2))(outputs)
    normalization_layer = cross_channel_normalization_2d()
    outputs = normalization_layer(outputs)

    outputs = ZeroPadding2D(padding=(2, 2))(outputs)

    convolution_layer = Conv2D(filters=128,
                               kernel_size=(5, 5),
                               padding='same')(outputs)
    lambda_layers_conv2 = [convolution_layer]
    for i in range(2):
        split_layer = split_tensor_2d(axis=3, ratio_split=2, id_split=i)(outputs)
        lambda_layers_conv2.append(split_layer)
    outputs = Concatenate()(lambda_layers_conv2)

    # Conv3
    outputs = MaxPooling2D(pool_size=(3, 3),
                           strides=(2, 2))(outputs)
    normalization_layer = cross_channel_normalization_2d()
    outputs = normalization_layer(outputs)

    outputs = ZeroPadding2D(padding=(2, 2))(outputs)
    outputs = Conv2D(filters=384,
                     kernel_size=(3, 3),
                     padding='same')(outputs)

    # Conv4
    outputs = ZeroPadding2D(padding=(2, 2))(outputs)
    convolution_layer = Conv2D(filters=192,
                               kernel_size=(3, 3),
                               padding='same')(outputs)
    lambda_layers_conv4 = [convolution_layer]
    for i in range(2):
        split_layer = split_tensor_2d(axis=3, ratio_split=2, id_split=i)(outputs)
        lambda_layers_conv4.append(split_layer)
    outputs = Concatenate()(lambda_layers_conv4)

    # Conv5
    outputs = ZeroPadding2D(padding=(2, 2))(outputs)
    normalization_layer = cross_channel_normalization_2d()
    outputs = normalization_layer(outputs)

    convolution_layer = Conv2D(filters=128,
                               kernel_size=(3, 3),
                               padding='same')(outputs)
    lambda_layers_conv5 = [convolution_layer]
    for i in range(2):
        split_layer = split_tensor_2d(axis=3, ratio_split=2, id_split=i)(outputs)
        lambda_layers_conv5.append(split_layer)
    outputs = Concatenate()(lambda_layers_conv5)

    outputs = MaxPooling2D(pool_size=(3, 3),
                           strides=(2, 2))(outputs)
    outputs = Flatten()(outputs)
    outputs = Dense(units=number_of_dense_units,
                    activation ='relu')(outputs)
    if dropout_rate > 0.0:
        outputs = Dropout(rate=dropout_rate)(outputs)
    outputs = Dense(units=number_of_dense_units,
                    activation ='relu')(outputs)
    if dropout_rate > 0.0:
        outputs = Dropout(rate=dropout_rate)(outputs)

    layer_activation = ''
    if mode == 'classification':
        layer_activation = 'softmax'
    elif mode == 'regression':
        layerActivation = 'linear'
    else:
        raise ValueError('unrecognized mode.')

    outputs = Dense(units=number_of_classification_labels,
                    activation=layer_activation)(outputs)

    alexnet_model = Model(inputs=inputs, outputs=outputs)

    return(alexnet_model)


def create_alexnet_model_3d(input_image_size,
                            number_of_classification_labels=1000,
                            number_of_dense_units=4096,
                            dropout_rate=0.0,
                            mode='classification'):
    """
    3-D implementation of the AlexNet deep learning architecture.

    Creates a keras model of the AlexNet deep learning architecture for image
    recognition based on the paper

    A. Krizhevsky, and I. Sutskever, and G. Hinton. ImageNet Classification with Deep Convolutional Neural Networks.

    available here:

            http://papers.nips.cc/paper/4824-imagenet-classification-with-deep-convolutional-neural-networks.pdf

    This particular implementation was influenced by the following python
    implementation:

            https://github.com/duggalrahul/AlexNet-Experiments-Keras/
            https://github.com/lunardog/convnets-keras/

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The
        shape (or dimension) of that tensor is the image dimensions followed by
        the number of channels (e.g., red, green, and blue).  The batch size
        (i.e., number of training images) is not specified a priori.

    number_of_classification_labels : integer
        Number of segmentation labels.

    number_of_dense_units : integer
       Number of dense units.

    dropout_rate : scalar
       Optional regularization parameter between [0, 1]. Default = 0.0.

    mode : string
       'classification' or 'regression'.  Default = 'classification'.

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_alexnet_model_3d((128, 128, 128, 1))
    >>> model.summary()

    """

    def split_tensor_3d(axis=1, ratio_split=1, id_split=0):

        def f(X):
           Xdims = K.int_shape(X)
           div = Xdims[axis] // ratio_split
           axis_split = slice(id_split * div, (id_split + 1) * div)

           if axis == 0:
               output = X[axis_split,:, :, :, :]
           elif axis == 1:
               output = X[:,axis_split, :, :, :]
           elif axis == 2:
               output = X[:,:, axis_split, :, :]
           elif axis == 3:
               output = X[:,:, :, axis_split, :]
           elif axis == 4:
               output = X[:,:, :, :, axis_split]
           else:
               raise ValueError("axis specified is out of range.")

           return(output)

        def g(input_shape):
            output_shape = list(input_shape)
            output_shape[axis] = output_shape[axis] // ratio_split
            return(tuple(output_shape))

        return(Lambda(f, output_shape=lambda input_shape:g(input_shape)))

    def cross_channel_normalization_3d(alpha=1e-4, k=2, beta=0.75, n=5):

        def normalize_tensor_3d(X):

            X2 = K.square(X)

            half = n // 2

            extra_channels = K.spatial_3d_padding(
                K.permute_dimensions(X2, (1, 2, 3, 4, 0)),
                padding=((0, 0), (0, 0), (half, half)))
            extra_channels = K.permute_dimensions(
                extra_channels, (4, 0, 1, 2, 3))

            Xdims = K.int_shape(X)
            number_of_channels = int(Xdims[-1])

            scale = k
            for i in range(n):
                scale += alpha * extra_channels[:,:,:,:,i:(i + number_of_channels)]
            scale = scale ** beta

            return(X / scale)

        return(Lambda(normalize_tensor_3d, output_shape = lambda input_shape:input_shape))

    inputs = Input(shape = input_image_size)

    # Conv1
    outputs = Conv3D(filters = 96,
                     kernel_size=(11, 11, 11),
                     strides=(4, 4, 4),
                     activation='relu')(inputs)

    # Conv2
    outputs = MaxPooling3D(pool_size=(3, 3, 3),
                           strides=(2, 2, 2))(outputs)
    normalization_layer = cross_channel_normalization_3d()
    outputs = normalization_layer(outputs)

    outputs = ZeroPadding3D(padding=(2, 2, 2))(outputs)

    convolution_layer = Conv3D(filters=128,
                               kernel_size=(5, 5, 5),
                               padding='same')(outputs)
    lambda_layers_conv2 = [convolution_layer]
    for i in range(2):
        split_layer = split_tensor_3d(axis=4, ratio_split=2, id_split=i)(outputs)
        lambda_layers_conv2.append(split_layer)
    outputs = Concatenate()(lambda_layers_conv2)

    # Conv3
    outputs = MaxPooling3D(pool_size=(3, 3, 3),
                           strides=(2, 2, 2))(outputs)
    normalization_layer = cross_channel_normalization_3d()
    outputs = normalization_layer(outputs)

    outputs = ZeroPadding3D(padding=(2, 2, 2))(outputs)
    outputs = Conv3D(filters=384,
                     kernel_size=(3, 3, 3),
                     padding='same')(outputs)

    # Conv4
    outputs = ZeroPadding3D(padding=(2, 2, 2))(outputs)
    convolution_layer = Conv3D(filters=192,
                               kernel_size=(3, 3, 3),
                               padding='same')(outputs)
    lambda_layers_conv4 = [convolution_layer]
    for i in range(2):
        split_layer = split_tensor_3d(axis=4, ratio_split=2, id_split=i)(outputs)
        lambda_layers_conv4.append(split_layer)
    outputs = Concatenate()(lambda_layers_conv4)

    # Conv5
    outputs = ZeroPadding3D(padding=(2, 2, 2))(outputs)
    normalization_layer = cross_channel_normalization_3d()
    outputs = normalization_layer(outputs)

    convolution_layer = Conv3D(filters=128,
                               kernel_size=(3, 3, 3),
                               padding='same')(outputs)
    lambda_layers_conv5 = [convolution_layer]
    for i in range(2):
        split_layer = split_tensor_3d(axis=4, ratio_split=2, id_split=i)(outputs)
        lambda_layers_conv5.append(split_layer)
    outputs = Concatenate()(lambda_layers_conv5)

    outputs = MaxPooling3D(pool_size=(3, 3, 3),
                           strides=(2, 2, 2))(outputs)
    outputs = Flatten()(outputs)
    outputs = Dense(units=number_of_dense_units,
                    activation ='relu')(outputs)
    if dropout_rate > 0.0:
        outputs = Dropout(rate=dropout_rate)(outputs)
    outputs = Dense(units=number_of_dense_units,
                    activation ='relu')(outputs)
    if dropout_rate > 0.0:
        outputs = Dropout(rate=dropout_rate)(outputs)

    layer_activation = ''
    if mode == 'classification':
        layer_activation = 'softmax'
    elif mode == 'regression':
        layerActivation = 'linear'
    else:
        raise ValueError('unrecognized mode.')

    outputs = Dense(units=number_of_classification_labels,
                    activation=layer_activation)(outputs)

    alexnet_model = Model(inputs=inputs, outputs=outputs)

    return(alexnet_model)




import numpy as np
import tensorflow as tf
import ants

def cortical_thickness(t1,
                       antsxnet_cache_directory=None,
                       verbose=False):

    """
    Perform KellyKapowski cortical thickness using deep_atropos for
    segmentation.  Description concerning implementaiton and evaluation:

    https://www.medrxiv.org/content/10.1101/2020.10.19.20215392v1

    Arguments
    ---------
    t1 : ANTsImage
        input 3-D unprocessed T1-weighted brain image.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    Cortical thickness image and segmentation probability images.

    Example
    -------
    >>> image = ants.image_read("t1w_image.nii.gz")
    >>> kk = cortical_thickness(image)
    """

    from ..utilities import deep_atropos

    if t1.dimension != 3:
        raise ValueError("Image dimension must be 3.")

    atropos = deep_atropos(t1, do_preprocessing=True,
        antsxnet_cache_directory=antsxnet_cache_directory, verbose=True)

    # Kelly Kapowski cortical thickness

    kk_segmentation = ants.image_clone(atropos['segmentation_image'])
    kk_segmentation[kk_segmentation == 4] = 3
    gray_matter = atropos['probability_images'][2]
    white_matter = (atropos['probability_images'][3] + atropos['probability_images'][4])
    kk = ants.kelly_kapowski(s=kk_segmentation, g=gray_matter, w=white_matter,
                            its=45, r=0.025, m=1.5, x=0, verbose=int(verbose))

    return_dict = {'thickness_image' : kk,
                   'segmentation_image' : atropos['segmentation_image'],
                   'csf_probability_image' : atropos['probability_images'][1],
                   'gray_matter_probability_image' : atropos['probability_images'][2],
                   'white_matter_probability_image' : atropos['probability_images'][3],
                   'deep_gray_matter_probability_image' : atropos['probability_images'][4],
                   'brain_stem_probability_image' : atropos['probability_images'][5],
                   'cerebellum_probability_image' : atropos['probability_images'][6]
                  }
    return(return_dict)


def longitudinal_cortical_thickness(t1s,
                                    initial_template="oasis",
                                    number_of_iterations=1,
                                    refinement_transform="antsRegistrationSyNQuick[a]",
                                    antsxnet_cache_directory=None,
                                    verbose=False):

    """
    Perform KellyKapowski cortical thickness longitudinally using \code{deepAtropos}
    for segmentation of the derived single-subject template.  It takes inspiration from
    the work described here:

    https://pubmed.ncbi.nlm.nih.gov/31356207/

    Arguments
    ---------
    t1s : list of ANTsImage
        Input list of 3-D unprocessed t1-weighted brain images from a single subject.

    initial_template : string or ANTsImage
        Input image to define the orientation of the SST.  Can be a string (see
        get_antsxnet_data) or a specified template.  This allows the user to create a
        SST outside of this routine.

    number_of_iterations : int
        Defines the number of iterations for refining the SST.

    refinement_transform : string
       Transform for defining the refinement registration transform. See options in
       ants.registration.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    Cortical thickness image and segmentation probability images.

    Example
    -------
    >>> t1s = list()
    >>> t1s.append(ants.image_read("t1w_image.nii.gz"))
    >>> kk = longitudinal_cortical_thickness(image)
    """

    from ..utilities import get_antsxnet_data
    from ..utilities import preprocess_brain_image
    from ..utilities import deep_atropos

    ###################
    #
    #  Initial SST + optional affine refinement
    #
    ##################

    sst = None
    if isinstance(initial_template, str):
        template_file_name_path = get_antsxnet_data(initial_template, antsxnet_cache_directory=antsxnet_cache_directory)
        sst = ants.image_read(template_file_name_path)
    else:
        sst = initial_template

    for s in range(number_of_iterations):
        if verbose:
            print("Refinement iteration", s, "( out of", number_of_iterations, ")")

        sst_tmp = ants.image_clone(sst) * 0
        for i in range(len(t1s)):
            if verbose:
                print("***************************")
                print( "SST processing image", i, "( out of", len(t1s), ")")
                print( "***************************" )
            transform_type = "antsRegistrationSyNQuick[r]"
            if s > 0:
                transform_type = refinement_transform
            t1_preprocessed = preprocess_brain_image(t1s[i],
                truncate_intensity=(0.01, 0.99),
                brain_extraction_modality=None,
                template=sst,
                template_transform_type=transform_type,
                do_bias_correction=False,
                do_denoising=False,
                intensity_normalization_type="01",
                antsxnet_cache_directory=antsxnet_cache_directory,
                verbose=verbose)
            sst_tmp += t1_preprocessed['preprocessed_image']

        sst = sst_tmp / len(t1s)

    ###################
    #
    #  Preprocessing and affine transform to final SST
    #
    ##################

    t1s_preprocessed = list()
    for i in range(len(t1s)):
        if verbose:
            print("***************************")
            print( "Final processing image", i, "( out of", len(t1s), ")")
            print( "***************************" )
        t1_preprocessed = preprocess_brain_image(t1s[i],
            truncate_intensity=(0.01, 0.99),
            brain_extraction_modality="t1",
            template=sst,
            template_transform_type="antsRegistrationSyNQuick[a]",
            do_bias_correction=True,
            do_denoising=True,
            intensity_normalization_type="01",
            antsxnet_cache_directory=antsxnet_cache_directory,
            verbose=verbose)
        t1s_preprocessed.append(t1_preprocessed)

    ###################
    #
    #  Deep  Atropos of SST for priors
    #
    ##################

    sst_atropos = deep_atropos(sst, do_preprocessing=True,
        antsxnet_cache_directory=antsxnet_cache_directory, verbose=verbose)

    ###################
    #
    #  Traditional Atropos + KK for each iamge
    #
    ##################

    return_list = list()
    for i in range(len(t1s_preprocessed)):
        if verbose:
            print("Atropos for image", i, "( out of", len(t1s), ")")
        atropos_output = ants.atropos(t1s_preprocessed[i]['preprocessed_image'],
            x=t1s_preprocessed[i]['brain_mask'], i=sst_atropos['probability_images'][1:7],
            m="[0.1,1x1x1]", c="[5,0]", priorweight=0.5, p="Socrates[1]", verbose=int(verbose))

        kk_segmentation = ants.image_clone(atropos_output['segmentation'])
        kk_segmentation[kk_segmentation == 4] = 3
        gray_matter = atropos_output['probabilityimages'][1]
        white_matter = atropos_output['probabilityimages'][2] + atropos_output['probabilityimages'][3]
        kk = ants.kelly_kapowski(s=kk_segmentation, g=gray_matter, w=white_matter,
            its=45, r=0.025, m=1.5, x=0, verbose=int(verbose))

        t1_dict = {'preprocessed_image' : t1s_preprocessed[i]['preprocessed_image'],
                   'thickness_image' : kk,
                   'segmentation_image' : atropos_output['segmentation'],
                   'csf_probability_image' : atropos_output['probabilityimages'][0],
                   'gray_matter_probability_image' : atropos_output['probabilityimages'][1],
                   'white_matter_probability_image' : atropos_output['probabilityimages'][2],
                   'deep_gray_matter_probability_image' : atropos_output['probabilityimages'][3],
                   'brain_stem_probability_image' : atropos_output['probabilityimages'][4],
                   'cerebellum_probability_image' : atropos_output['probabilityimages'][5],
                   'template_transforms' : t1s_preprocessed[i]['template_transforms']
                  }
        return_list.append(t1_dict)

    return_list.append(sst)

    return(return_list)





from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Conv2D, Conv2DTranspose, Dense, Flatten, Reshape

import numpy as np
import math

def create_convolutional_autoencoder_model_2d(input_image_size,
                                              number_of_filters_per_layer=(32, 64, 128, 10),
                                              convolution_kernel_size=(5, 5),
                                              deconvolution_kernel_size=(5, 5)
                                             ):
    """
    Function for creating a 2-D symmetric convolutional autoencoder model.

    Builds an autoencoder based on the specified array definining the
    number of units in the encoding branch.  Ported from the Keras python
    implementation here:

    https://github.com/XifengGuo/DEC-keras

    Arguments
    ---------
    input_image_size : tuple
        A tuple defining the shape of the 2-D input image

    number_of_units_per_layer : tuple
        A tuple defining the number of units in the encoding branch.

    convolution_kernel_size : tuple or scalar
        Kernel size for convolution

    deconvolution_kernel_size : tuple or scalar
        Kernel size for deconvolution

    Returns
    -------
    Keras models
        A convolutional encoder and autoencoder Keras model.

    Example
    -------
    >>> autoencoder, encoder = create_convolutional_autoencoder_model_2d((128, 128, 3))
    >>> autoencoder.summary()
    >>> encoder.summary()
    """

    activation = 'relu'
    strides = (2, 2)

    number_of_encoding_layers = len(number_of_filters_per_layer) - 1

    factor = 2 ** number_of_encoding_layers

    padding = 'valid'
    if input_image_size[0] % factor == 0:
        padding = 'same'

    inputs = Input(shape = input_image_size)

    encoder = inputs

    for i in range(number_of_encoding_layers):
        local_padding = 'same'
        kernel_size = convolution_kernel_size
        if i == (number_of_encoding_layers - 1):
            local_padding = padding
            kernel_size = tuple(np.array(convolution_kernel_size) - 2)

        encoder = Conv2D(filters=number_of_filters_per_layer[i],
                         kernel_size=kernel_size,
                         strides=strides,
                         activation=activation,
                         padding=local_padding)(encoder)

    encoder = Flatten()(encoder)
    encoder = Dense(units=number_of_filters_per_layer[-1])(encoder)

    autoencoder = encoder

    penultimate_number_of_filters = \
      number_of_filters_per_layer[number_of_encoding_layers-1]

    input_image_size_factored = ((math.floor(input_image_size[0] / factor)),
                                 (math.floor(input_image_size[1] / factor)))

    number_of_units_for_encoder_output = (penultimate_number_of_filters *
      input_image_size_factored[0] * input_image_size_factored[1])

    autoencoder = Dense(units=number_of_units_for_encoder_output,
                        activation=activation)(autoencoder)
    autoencoder = Reshape(target_shape=(*input_image_size_factored, penultimate_number_of_filters))(autoencoder)

    for i in range(number_of_encoding_layers, 1, -1):
        local_padding = 'same'
        kernel_size = convolution_kernel_size
        if i == number_of_encoding_layers:
            local_padding = padding
            kernel_size = tuple(np.array(deconvolution_kernel_size) - 2)

        autoencoder = Conv2DTranspose(filters=number_of_filters_per_layer[i-2],
                                      kernel_size=kernel_size,
                                      strides=strides,
                                      activation=activation,
                                      padding=local_padding)(autoencoder)

    autoencoder = Conv2DTranspose(input_image_size[-1],
                                  kernel_size=deconvolution_kernel_size,
                                  strides=strides,
                                  padding='same')(autoencoder)

    autoencoder_model = Model(inputs=inputs, outputs=autoencoder)
    encoder_model = Model(inputs=inputs, outputs=encoder)

    return([autoencoder_model, encoder_model])


def create_convolutional_autoencoder_model_3d(input_image_size,
                                              number_of_filters_per_layer=(32, 64, 128, 10),
                                              convolution_kernel_size=(5, 5, 5),
                                              deconvolution_kernel_size=(5, 5, 5)
                                             ):
    """
    Function for creating a 3-D symmetric convolutional autoencoder model.

    Builds an autoencoder based on the specified array definining the
    number of units in the encoding branch.  Ported from the Keras python
    implementation here:

    https://github.com/XifengGuo/DEC-keras

    Arguments
    ---------
    input_image_size : tuple
        A tuple defining the shape of the 3-D input image

    number_of_units_per_layer : tuple
        A tuple defining the number of units in the encoding branch.

    convolution_kernel_size : tuple or scalar
        Kernel size for convolution

    deconvolution_kernel_size : tuple or scalar
        Kernel size for deconvolution

    Returns
    -------
    Keras models
        A convolutional encoder and autoencoder Keras model.

    Example
    -------
    >>> autoencoder, encoder = create_convolutional_autoencoder_model_3d((128, 128, 128, 3))
    >>> autoencoder.summary()
    >>> encoder.summary()
    """

    activation = 'relu'
    strides = (2, 2, 2)

    number_of_encoding_layers = len(number_of_filters_per_layer) - 1

    factor = 2 ** number_of_encoding_layers

    padding = 'valid'
    if input_image_size[0] % factor == 0:
        padding = 'same'

    inputs = Input(shape = input_image_size)

    encoder = inputs

    for i in range(number_of_encoding_layers):
        local_padding = 'same'
        kernel_size = convolution_kernel_size
        if i == (number_of_encoding_layers - 1):
            local_padding = padding
            kernel_size = tuple(np.array(convolution_kernel_size) - 2)

        encoder = Conv3D(filters=number_of_filters_per_layer[i],
                         kernel_size=kernel_size,
                         strides=strides,
                         activation=activation,
                         padding=local_padding)(encoder)

    encoder = Flatten()(encoder)
    encoder = Dense(units=number_of_filters_per_layer[-1])(encoder)

    autoencoder = encoder

    penultimate_number_of_filters = \
      number_of_filters_per_layer[number_of_encoding_layers-1]

    input_image_size_factored = ((math.floor(input_image_size[0] / factor)),
                                 (math.floor(input_image_size[1] / factor)),
                                 (math.floor(input_image_size[2] / factor)))

    number_of_units_for_encoder_output = (penultimate_number_of_filters *
      input_image_size_factored[0] * input_image_size_factored[1] *
      input_image_size_factored[2])

    autoencoder = Dense(units=number_of_units_for_encoder_output,
                        activation=activation)(autoencoder)
    autoencoder = Reshape(target_shape=(*input_image_size_factored, penultimate_number_of_filters))(autoencoder)

    for i in range(number_of_encoding_layers, 1, -1):
        local_padding = 'same'
        kernel_size = convolution_kernel_size
        if i == number_of_encoding_layers:
            local_padding = padding
            kernel_size = tuple(np.array(deconvolution_kernel_size) - 2)

        autoencoder = Conv3DTranspose(filters=number_of_filters_per_layer[i-2],
                                      kernel_size=kernel_size,
                                      strides=strides,
                                      activation=activation,
                                      padding=local_padding)(autoencoder)

    autoencoder = Conv3DTranspose(input_image_size[-1],
                                  kernel_size=deconvolution_kernel_size,
                                  strides=strides,
                                  padding='same')(autoencoder)

    autoencoder_model = Model(inputs=inputs, outputs=autoencoder)
    encoder_model = Model(inputs=inputs, outputs=encoder)

    return([autoencoder_model, encoder_model])




import numpy as np
import ants

def hypothalamus_segmentation(t1,
                              antsxnet_cache_directory=None,
                              verbose=False):

    """
    Hypothalamus and subunits segmentation

    Described here:

        https://pubmed.ncbi.nlm.nih.gov/32853816/

    ported from the original implementation

        https://github.com/BBillot/hypothalamus_seg

    Subunits labeling:

    Label 1:  left anterior-inferior
    Label 2:  left anterior-superior
    Label 3:  left posterior
    Label 4:  left tubular inferior
    Label 5:  left tubular superior
    Label 6:  right anterior-inferior
    Label 7:  right anterior-superior
    Label 8:  right posterior
    Label 9:  right tubular inferior
    Label 10: right tubular superior

    Arguments
    ---------
    t1 : ANTsImage
        input 3-D T1 brain image.

    antsxnet_cache_directory : string
        Destination directory for storing the downloaded template and model weights.
        Since these can be resused, if is None, these data will be downloaded to a
        ~/.keras/ANTsXNet/.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    Hypothalamus segmentation (and subunits) probability images

    Example
    -------
    >>> image = ants.image_read("t1.nii.gz")
    >>> hypo = hypothalamus_segmentation(image)
    """

    from ..architectures import create_hypothalamus_unet_model_3d
    from ..utilities import get_pretrained_network

    if t1.dimension != 3:
        raise ValueError( "Image dimension must be 3." )

    if antsxnet_cache_directory == None:
        antsxnet_cache_directory = "ANTsXNet"

    classes = ("background",
               "left anterior-inferior",
               "left anterior-superior",
               "left posterior",
               "left tubular inferior",
               "left tubular superior",
               "right anterior-inferior",
               "right anterior-superior",
               "right posterior",
               "right tubular inferior",
               "right tubular superior")

    ################################
    #
    # Rotate to proper orientation
    #
    ################################

    reference_image = ants.make_image((256, 256, 256),
                                      voxval=0,
                                      spacing=(1, 1, 1),
                                      origin=(0, 0, 0),
                                      direction=np.diag((-1.0, -1.0, 1.0)))
    center_of_mass_reference = ants.get_center_of_mass(reference_image + 1)
    center_of_mass_image = ants.get_center_of_mass(t1 * 0 + 1)
    translation = np.asarray(center_of_mass_image) - np.asarray(center_of_mass_reference)
    xfrm = ants.create_ants_transform(transform_type="Euler3DTransform",
        center=np.asarray(center_of_mass_reference), translation=translation)
    xfrm_inv = xfrm.invert()

    crop_image = ants.image_clone(t1) * 0 + 1
    crop_image = ants.apply_ants_transform_to_image(xfrm, crop_image, reference_image)
    crop_image = ants.crop_image(crop_image, label_image=crop_image, label=1)

    t1_warped = ants.apply_ants_transform_to_image(xfrm, t1, crop_image)

    ################################
    #
    # Normalize intensity
    #
    ################################

    t1_warped = ((t1_warped - t1_warped.min()) / (t1_warped.max() - t1_warped.min()))

    ################################
    #
    # Build models and load weights
    #
    ################################

    if verbose == True:
        print("Hypothalamus:  retrieving model weights.")

    unet_model = create_hypothalamus_unet_model_3d(t1_warped.shape)

    weights_file_name = get_pretrained_network("hypothalamus", antsxnet_cache_directory=antsxnet_cache_directory)
    unet_model.load_weights(weights_file_name)

    ################################
    #
    # Do prediction
    #
    ################################

    if verbose == True:
        print("Prediction.")

    batchX = np.zeros((1, *t1_warped.shape, 1))
    batchX[0,:,:,:,0] = t1_warped.numpy()

    predicted_data = unet_model.predict(batchX, verbose=verbose)

    probability_images = list()
    for i in range(len(classes)):
        if verbose == True:
            print("Processing image", classes[i])

        probability_image = ants.from_numpy(np.squeeze(predicted_data[0,:,:,:,i]),
            spacing=t1_warped.spacing, origin=t1_warped.origin,
            direction=t1_warped.direction)
        probability_images.append(xfrm_inv.apply_to_image(probability_image, t1))

    image_matrix = ants.image_list_to_matrix(probability_images, t1 * 0 + 1)
    segmentation_matrix = np.argmax(image_matrix, axis=0)
    segmentation_image = ants.matrix_to_images(
        np.expand_dims(segmentation_matrix, axis=0), t1 * 0 + 1)[0]

    return_dict = {'segmentation_image' : segmentation_image,
                   'probability_images' : probability_images}
    return(return_dict)






import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model
from tensorflow.keras.layers import (Input, Dropout, BatchNormalization, Add,
                                     Concatenate, Dense, Activation,
                                     Conv2D, Conv2DTranspose, AveragePooling2D,
                                     MaxPooling2D, UpSampling2D, ZeroPadding2D,
                                     Conv3D, Conv3DTranspose, AveragePooling3D,
                                     MaxPooling3D, UpSampling3D, ZeroPadding3D)

from tensorflow.keras import initializers
from tensorflow.keras import regularizers

from ..utilities import Scale

def create_denseunet_model_2d(input_image_size,
                              number_of_outputs=1,
                              number_of_layers_per_dense_block=(6, 12, 36, 24),
                              growth_rate=48,
                              initial_number_of_filters=96,
                              reduction_rate=0.0,
                              depth=7,
                              dropout_rate=0.0,
                              weight_decay=1e-4,
                              mode='classification'
                             ):
    """
    2-D implementation of the dense U-net deep learning architecture.

    Creates a keras model of the dense U-net deep learning architecture for
    image segmentation

    X. Li, H. Chen, X. Qi, Q. Dou, C.-W. Fu, P.-A. Heng. H-DenseUNet: Hybrid
    Densely Connected UNet for Liver and Tumor Segmentation from CT Volumes

    available here:

            https://arxiv.org/pdf/1709.07330.pdf

    with the author's implementation available at:

            https://github.com/xmengli999/H-DenseUNet

    Arguments
    ---------
    input_image_size : tuple of length 3
        Used for specifying the input tensor shape.  The
        shape (or dimension) of that tensor is the image dimensions followed by
        the number of channels (e.g., red, green, and blue).  The batch size
        (i.e., number of training images) is not specified a priori.

    number_of_outputs : integer
        Meaning depends on the mode.  For 'classification' this is the number of
        segmentation labels.  For 'regression' this is the number of outputs.

    number_of_layers_per_dense_blocks : tuple
        Number of dense blocks per layer.

    growth_rate : integer
        Number of filters to add for each dense block layer (default = 48).

    initial_number_of_filters : integer
        Number of filters at the beginning (default = 96).

    reduction_rate : scalar
        Reduction factor of transition blocks.

    depth :  integer
        Number of layers---must be equal to 3 * N + 4 where N is an integer
        (default = 7).

    dropout_rate : scalar
        Float between 0 and 1 to use between dense layers.

    weight_decay :  scalar
        Weighting parameter for L2 regularization of the kernel weights of the
        convolution layers (default = 1e-4).

    Returns
    -------
    Keras model
        A 2-D Keras model defining the network.

    Example
    -------
    >>> model = create_denseunet_model_2d((128, 128, 1))
    >>> model.summary()
    """

    concatenation_axis=1
    if K.image_data_format() == 'channels_last':
        concatenation_axis=-1

    def convolution_factory_2d(model, number_of_filters,
                               kernel_size=(3, 3),
                               dropout_rate=0.0, weight_decay=1e-4):

        # Bottleneck layer

        model = BatchNormalization(axis=concatenation_axis)(model)
        model = Scale(axis=concatenation_axis)(model)
        model = Activation('relu')(model)
        model = Conv2D(filters=(number_of_filters * 4),
                       kernel_size=(1, 1),
                       use_bias=False)(model)

        if dropout_rate > 0.0:
            model = Dropout(rate=dropout_rate)(model)

        # Convolution layer

        model = BatchNormalization(axis=concatenation_axis,
                                   epsilon=1.1e-5)(model)
        model = Scale(axis=concatenation_axis)(model)
        model = Activation(activation='relu')(model)
        model = ZeroPadding2D(padding=(1, 1))(model)
        model = Conv2D(filters=number_of_filters,
                       kernel_size=kernel_size,
                       use_bias=False)(model)

        if dropout_rate > 0.0:
            model = Dropout(rate=dropout_rate)(model)

        return(model)

    def transition_2d(model, number_of_filters, compression_rate=1.0,
                      dropout_rate=0.0, weight_decay=1e-4):

        model = BatchNormalization(axis=concatenation_axis,
                                   gamma_regularizer=regularizers.l2(weight_decay),
                                   beta_regularizer=regularizers.l2(weight_decay))(model)
        model = Scale(axis=concatenation_axis)(model)
        model = Activation(activation='relu')(model)
        model = Conv2D(filters=int(number_of_filters * compression_rate),
                       kernel_size=(1, 1),
                       use_bias=False)(model)

        if dropout_rate > 0.0:
            model = Dropout(rate=dropout_rate)(model)

        model = AveragePooling2D(pool_size=(2, 2),
                                 strides=(2, 2))(model)
        return(model)

    def create_dense_blocks_2d(model, number_of_filters, depth, growth_rate,
                               dropout_rate=0.0, weight_decay=1e-4):

        dense_block_layers = [model]
        for i in range(depth):
            model = convolution_factory_2d(model, number_of_filters=growth_rate,
                                           kernel_size=(3, 3), dropout_rate=dropout_rate,
                                           weight_decay=weight_decay)
            dense_block_layers.append(model)
            model = Concatenate(axis=concatenation_axis)(dense_block_layers)
            number_of_filters += growth_rate

        return(model, number_of_filters)


    if ((depth - 4) % 3) != 0:
        raise ValueError('Depth must be equal to 3*N+4 where N is an integer.')

    number_of_layers = int((depth - 4) % 3)
    number_of_dense_blocks = len(number_of_layers_per_dense_block)

    inputs = Input(shape = input_image_size)

    box_layers = []
    box_count = 1

    # Initial convolution

    outputs = ZeroPadding2D(padding=(3, 3))(inputs)
    outputs = Conv2D(filters=initial_number_of_filters,
                     kernel_size=(7, 7),
                     strides=(2, 2),
                     use_bias=False)(outputs)
    outputs = BatchNormalization(epsilon=1.1e-5,
                                 axis=concatenation_axis)(outputs)
    outputs = Scale(axis=concatenation_axis)(outputs)
    outputs = Activation(activation='relu')(outputs)

    box_layers.append(outputs)
    box_count += 1

    outputs = ZeroPadding2D(padding=(1, 1))(outputs)
    outputs = MaxPooling2D(pool_size=(3, 3),
                           strides=(2, 2))(outputs)

    # Add dense blocks

    nFilters = initial_number_of_filters

    for i in range(number_of_dense_blocks - 1):
        outputs, number_of_filters = \
           create_dense_blocks_2d(outputs, number_of_filters=nFilters,
                                  depth=number_of_layers_per_dense_block[i],
                                  growth_rate=growth_rate, dropout_rate=dropout_rate,
                                  weight_decay=weight_decay)
        box_layers.append(outputs)
        box_count += 1

        outputs = transition_2d(outputs, number_of_filters=number_of_filters,
                                compression_rate=(1.0 - reduction_rate),
                                dropout_rate=dropout_rate, weight_decay=weight_decay)
        nFilters = int(number_of_filters * (1 - reduction_rate))


    outputs, nFilters = \
       create_dense_blocks_2d(outputs, number_of_filters=nFilters,
                              depth=number_of_layers_per_dense_block[number_of_dense_blocks - 1],
                              growth_rate=growth_rate, dropout_rate=dropout_rate,
                              weight_decay=weight_decay)

    outputs = BatchNormalization(epsilon=1.1e-5,
                                 axis=concatenation_axis)(outputs)
    outputs = Scale(axis=concatenation_axis)(outputs)
    outputs = Activation(activation='relu')(outputs)

    box_layers.append(outputs)
    box_count -= 1

    local_number_of_filters = (K.int_shape(box_layers[box_count]))[-1]
    local_layer = Conv2D(filters=local_number_of_filters,
                         kernel_size=(1, 1),
                         padding='same',
                         kernel_initializer='normal')(box_layers[box_count - 1])
    box_count -= 1

    for i in range(number_of_dense_blocks - 1):
        upsampling_layer = UpSampling2D(size=(2, 2))(outputs)
        outputs = Add()([local_layer, upsampling_layer])

        local_layer = box_layers[box_count - 1]
        box_count -= 1

        local_number_of_filters = (K.int_shape(box_layers[box_count]))[-1]
        outputs = Conv2D(filters=local_number_of_filters,
                         kernel_size=(3, 3),
                         padding='same',
                         kernel_initializer='normal')(outputs)

        if i == (number_of_dense_blocks - 2):
            outputs = Dropout(rate=0.3)(outputs)

        outputs = BatchNormalization(epsilon=1.1e-5,
                                     axis=concatenation_axis)(outputs)
        outputs = Activation(activation='relu')(outputs)

    convActivation = ''

    if mode == 'classification':
        convActivation = 'softmax'
    elif mode == 'regression':
        convActivation = 'linear'
    else:
        raise ValueError('mode must be either `classification` or `regression`.')

    outputs = Conv2D(filters=number_of_outputs,
                     kernel_size=(1, 1),
                     activation = convActivation,
                     kernel_initializer='normal')(outputs)

    denseunet_model = Model(inputs=inputs, outputs=outputs)

    return denseunet_model


def create_denseunet_model_3d(input_image_size,
                              number_of_outputs=1,
                              number_of_layers_per_dense_block=(6, 12, 36, 24),
                              growth_rate=48,
                              initial_number_of_filters=96,
                              reduction_rate=0.0,
                              depth=7,
                              dropout_rate=0.0,
                              weight_decay=1e-4,
                              mode='classification'
                             ):
    """
    2-D implementation of the dense U-net deep learning architecture.

    Creates a keras model of the dense U-net deep learning architecture for
    image segmentation

    X. Li, H. Chen, X. Qi, Q. Dou, C.-W. Fu, P.-A. Heng. H-DenseUNet: Hybrid
    Densely Connected UNet for Liver and Tumor Segmentation from CT Volumes

    available here:

            https://arxiv.org/pdf/1709.07330.pdf

    with the author's implementation available at:

            https://github.com/xmengli999/H-DenseUNet

    Arguments
    ---------
    input_image_size : tuple of length 4
        Used for specifying the input tensor shape.  The
        shape (or dimension) of that tensor is the image dimensions followed by
        the number of channels (e.g., red, green, and blue).  The batch size
        (i.e., number of training images) is not specified a priori.

    number_of_outputs : integer
        Meaning depends on the mode.  For 'classification' this is the number of
        segmentation labels.  For 'regression' this is the number of outputs.

    number_of_layers_per_dense_blocks : tuple
        Number of dense blocks per layer.

    growth_rate : integer
        Number of filters to add for each dense block layer (default = 48).

    initial_number_of_filters : integer
        Number of filters at the beginning (default = 96).

    reduction_rate : scalar
        Reduction factor of transition blocks.

    depth :  integer
        Number of layers---must be equal to 3 * N + 4 where N is an integer
        (default = 7).

    dropout_rate : scalar
        Float between 0 and 1 to use between dense layers.

    weight_decay :  scalar
        Weighting parameter for L2 regularization of the kernel weights of the
        convolution layers (default = 1e-4).

    Returns
    -------
    Keras model
        A 3-D Keras model defining the network.

    Example
    -------
    >>> model = create_denseunet_model_3d((128, 128, 128, 1))
    >>> model.summary()
    """

    concatenation_axis=1
    if K.image_data_format() == 'channels_last':
        concatenation_axis=-1

    def convolution_factory_3d(model, number_of_filters,
                               kernel_size=(3, 3, 3),
                               dropout_rate=0.0, weight_decay=1e-4):

        # Bottleneck layer

        model = BatchNormalization(axis=concatenation_axis)(model)
        model = Scale(axis=concatenation_axis)(model)
        model = Activation('relu')(model)
        model = Conv3D(filters=(number_of_filters * 4),
                       kernel_size=(1, 1, 1),
                       use_bias=False)(model)

        if dropout_rate > 0.0:
            model = Dropout(rate=dropout_rate)(model)

        # Convolution layer

        model = BatchNormalization(axis=concatenation_axis,
                                   epsilon=1.1e-5)(model)
        model = Scale(axis=concatenation_axis)(model)
        model = Activation(activation='relu')(model)
        model = ZeroPadding3D(padding=(1, 1, 1))(model)
        model = Conv3D(filters=number_of_filters,
                       kernel_size=kernel_size,
                       use_bias=False)(model)

        if dropout_rate > 0.0:
            model = Dropout(rate=dropout_rate)(model)

        return(model)

    def transition_3d(model, number_of_filters, compression_rate=1.0,
                      dropout_rate=0.0, weight_decay=1e-4):

        model = BatchNormalization(axis=concatenation_axis,
                                   gamma_regularizer=regularizers.l2(weight_decay),
                                   beta_regularizer=regularizers.l2(weight_decay))(model)
        model = Scale(axis=concatenation_axis)(model)
        model = Activation(activation='relu')(model)
        model = Conv3D(filters=int(number_of_filters * compression_rate),
                       kernel_size=(1, 1, 1),
                       use_bias=False)(model)

        if dropout_rate > 0.0:
            model = Dropout(rate=dropout_rate)(model)

        model = AveragePooling3D(pool_size=(2, 2, 2),
                                 strides=(2, 2, 2))(model)
        return(model)

    def create_dense_blocks_3d(model, number_of_filters, depth, growth_rate,
                               dropout_rate=0.0, weight_decay=1e-4):

        dense_block_layers = [model]
        for i in range(depth):
            model = convolution_factory_3d(model, number_of_filters=growth_rate,
                                           kernel_size=(3, 3, 3), dropout_rate=dropout_rate,
                                           weight_decay=weight_decay)
            dense_block_layers.append(model)
            model = Concatenate(axis=concatenation_axis)(dense_block_layers)
            number_of_filters += growth_rate

        return(model, number_of_filters)


    if ((depth - 4) % 3) != 0:
        raise ValueError('Depth must be equal to 3*N+4 where N is an integer.')

    number_of_layers = int((depth - 4) % 3)
    number_of_dense_blocks = len(number_of_layers_per_dense_block)

    inputs = Input(shape = input_image_size)

    box_layers = []
    box_count = 1

    # Initial convolution

    outputs = ZeroPadding3D(padding=(3, 3))(inputs)
    outputs = Conv3D(filters=initial_number_of_filters,
                     kernel_size=(7, 7, 7),
                     strides=(2, 2, 2),
                     use_bias=False)(outputs)
    outputs = BatchNormalization(epsilon=1.1e-5,
                                 axis=concatenation_axis)(outputs)
    outputs = Scale(axis=concatenation_axis)(outputs)
    outputs = Activation(activation='relu')(outputs)

    box_layers.append(outputs)
    box_count += 1

    outputs = ZeroPadding3D(padding=(1, 1, 1))(outputs)
    outputs = MaxPooling3D(pool_size=(3, 3, 3),
                           strides=(2, 2, 2))(outputs)

    # Add dense blocks

    nFilters = initial_number_of_filters

    for i in range(number_of_dense_blocks - 1):
        outputs, number_of_filters = \
           create_dense_blocks_3d(outputs, number_of_filters=nFilters,
                                  depth=number_of_layers_per_dense_block[i],
                                  growth_rate=growth_rate, dropout_rate=dropout_rate,
                                  weight_decay=weight_decay)
        box_layers.append(outputs)
        box_count += 1

        outputs = transition_3d(outputs, number_of_filters=number_of_filters,
                                compression_rate=(1.0 - reduction_rate),
                                dropout_rate=dropout_rate, weight_decay=weight_decay)
        nFilters = int(number_of_filters * (1 - reduction_rate))


    outputs, nFilters = \
       create_dense_blocks_3d(outputs, number_of_filters=nFilters,
                              depth=number_of_layers_per_dense_block[number_of_dense_blocks - 1],
                              growth_rate=growth_rate, dropout_rate=dropout_rate,
                              weight_decay=weight_decay)

    outputs = BatchNormalization(epsilon=1.1e-5,
                                 axis=concatenation_axis)(outputs)
    outputs = Scale(axis=concatenation_axis)(outputs)
    outputs = Activation(activation='relu')(outputs)

    box_layers.append(outputs)
    box_count -= 1

    local_number_of_filters = (K.int_shape(box_layers[box_count]))[-1]
    local_layer = Conv3D(filters=local_number_of_filters,
                         kernel_size=(1, 1, 1),
                         padding='same',
                         kernel_initializer='normal')(box_layers[box_count - 1])
    box_count -= 1

    for i in range(number_of_dense_blocks - 1):
        upsampling_layer = UpSampling3D(size=(2, 2, 2))(outputs)
        outputs = Add()([local_layer, upsampling_layer])

        local_layer = box_layers[box_count - 1]
        box_count -= 1

        local_number_of_filters = (K.int_shape(box_layers[box_count]))[-1]
        outputs = Conv3D(filters=local_number_of_filters,
                         kernel_size=(3, 3, 3),
                         padding='same',
                         kernel_initializer='normal')(outputs)

        if i == (number_of_dense_blocks - 2):
            outputs = Dropout(rate=0.3)(outputs)

        outputs = BatchNormalization(epsilon=1.1e-5,
                                     axis=concatenation_axis)(outputs)
        outputs = Activation(activation='relu')(outputs)

    convActivation = ''

    if mode == 'classification':
        convActivation = 'softmax'
    elif mode == 'regression':
        convActivation = 'linear'
    else:
        raise ValueError('mode must be either `classification` or `regression`.')

    outputs = Conv3D(filters=number_of_outputs,
                     kernel_size=(1, 1, 1),
                     activation = convActivation,
                     kernel_initializer='normal')(outputs)

    denseunet_model = Model(inputs=inputs, outputs=outputs)

    return denseunet_model






import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import (Input, Add, BatchNormalization,
                                     Conv2D, Conv3D, Dense, ReLU, LeakyReLU,
                                     UpSampling2D, UpSampling3D)
from tensorflow.keras.optimizers import Adam

from tensorflow.keras.applications import vgg19

from . import create_vgg_model_2d, create_vgg_model_3d

import numpy as np
import os

import matplotlib.pyplot as plot

import ants

class SuperResolutionGanModel(object):
    """
    Super resolution GAN model

    Super resolution generative adverserial network from the paper:

      https://arxiv.org/abs/1609.04802

    and ported from the Keras implementation:

      https://github.com/eriklindernoren/Keras-GAN/blob/master/wgan/wgan.py

    Arguments
    ---------
    input_image_size : tuple
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    low_resolution_image_size : tuple
        Size of the input image.

    scale_factor : integer
        Upsampling factor for the output super-resolution image.

    use_image_net_weights : boolean
        Determines whether or not one uses the image-net weights.  Only valid for
        2-D images.

    number_of_residual_blocks : 16
        Number of residual blocks used in constructing the generator.

    number_of_filters_at_base_layer : tuple of length 2
        Number of filters at the base layer for the generator and discriminator,
        respectively.

    Returns
    -------
    Keras model
        A Keras model defining the network.
    """

    def __init__(self, low_resolution_image_size, scale_factor=2,
                 use_image_net_weights=True, number_of_residual_blocks=16,
                 number_of_filters_at_base_layer=(64, 64)):
        super(SuperResolutionGanModel, self).__init__()

        self.low_resolution_image_size = low_resolution_image_size
        self.number_of_channels = self.low_resolution_image_size[-1]
        self.number_of_residual_blocks = number_of_residual_blocks
        self.number_of_filters_at_base_layer = number_of_filters_at_base_layer
        self.use_image_net_weights = use_image_net_weights

        self.scale_factor = scale_factor
        if not self.scale_factor in set([1, 2, 4, 8]):
            raise ValueError("Scale factor must be one of 1, 2, 4, or 8.")

        self.dimensionality = None
        if len(self.low_resolution_image_size) == 3:
            self.dimensionality = 2
        elif len(self.low_resolution_image_size) == 4:
            self.dimensionality = 3
            if self.use_image_net_weights == True:
                self.use_image_net_weights = False
            print("Warning:  imageNet weights are unavailable for 3D.")
        else:
            raise ValueError("Incorrect size for low_resolution_image_size.")

        optimizer = Adam(lr=0.0002, beta_1=0.5)

        # Images

        tmp = list(self.low_resolution_image_size)
        for i in range(self.dimensionality):
            tmp[i] *= self.scale_factor
        self.high_resolution_image_size = tuple(tmp)

        high_resolution_image = Input(shape=self.high_resolution_image_size)

        low_resolution_image = Input(shape=self.low_resolution_image_size)

        # Build generator

        self.generator = self.build_generator()

        fake_high_resolution_image = self.generator(low_resolution_image)

        # Build discriminator

        self.discriminator = self.build_discriminator()
        self.discriminator.compile(loss='mse',
                                   optimizer=optimizer, metrics=['acc'])

        # Vgg

        self.vgg_model = self.build_truncated_vgg_model()
        self.vgg_model.trainable = False
        self.vgg_model.compile(loss='mse', optimizer=optimizer, metrics=['acc'])

        if self.dimensionality == 2:
            self.discriminator_patch_size = (16, 16, 1)
        else:
            self.discriminator_patch_size = (16, 16, 16, 1)

        # Discriminator

        self.discriminator.trainable = False

        validity = self.discriminator(fake_high_resolution_image)

        # Combined model

        if self.use_image_net_weights == True:
            fake_features = self.vgg_model(fake_high_resolution_image)
            self.combined_model = Model(inputs=[low_resolution_image, high_resolution_image],
                                        outputs=[validity, fake_features])
            self.combined_model.compile(loss=['binary_crossentropy', 'mse'],
              loss_weights=[1e-3, 1], optimizer=optimizer)
        else:
            self.combined_model = Model(inputs=[low_resolution_image, high_resolution_image],
                                        outputs=validity)
            self.combined_model.compile(loss=['binary_crossentropy'], optimizer=optimizer)

    def build_truncated_vgg_model(self):
        vgg_tmp = None
        if self.dimensionality == 2:
            if self.use_image_net_weights == True:
                vgg_tmp = create_vgg_model_2d((224, 224, 3), style=19)
                keras_vgg = vgg19(weights='imagenet')
                vgg_tmp.set_weights(keras_vgg.get_weights())
            else:
                vgg_tmp = create_vgg_model_2d(
                  self.high_resolution_image_size, style=19)
        else:
            vgg_tmp = create_vgg_model_3d(self.high_resolution_image_size, style=19)

        vgg_tmp.outputs = [vgg_tmp.layers[9].output]

        high_resolution_image = Input(shape=self.high_resolution_image_size)
        high_resolution_image_features = vgg_tmp(high_resolution_image)

        vgg_model = Model(inputs=high_resolution_image,
                          outputs=high_resolution_image_features)
        return(vgg_model)

    def build_generator(self, number_of_filters=64):

        def build_residual_block(input, number_of_filters, kernel_size=3):
           shortcut = input
           if self.dimensionality == 2:
               input = Conv2D(filters=number_of_filters,
                              kernel_size=kernel_size,
                              strides=1,
                              padding='same')(input)
           else:
               input = Conv3D(filters=number_of_filters,
                              kernel_size=kernel_size,
                              strides=1,
                              padding='same')(input)
           input = ReLU()(input)
           input = BatchNormalization(momentum=0.8)(input)
           if self.dimensionality == 2:
               input = Conv2D(filters=number_of_filters,
                              kernel_size=kernel_size,
                              strides=1,
                              padding='same')(input)
           else:
               input = Conv3D(filters=number_of_filters,
                              kernel_size=kernel_size,
                              strides=1,
                              padding='same')(input)
           input = BatchNormalization(momentum=0.8)(input)
           input = Add()([input, shortcut])
           return(input)

        def build_deconvolution_layer(input, number_of_filters=256, kernel_size=3):
            model = input
            if self.dimensionality == 2:
                model = UpSampling2D(size=2)(model)
                input = Conv2D(filters=number_of_filters,
                               kernel_size=kernel_size,
                               strides=1,
                               padding='same')(model)
            else:
                model = UpSampling3D(size=2)(model)
                input = Conv3D(filters=number_of_filters,
                               kernel_size=kernel_size,
                               strides=1,
                               padding='same')(model)
            model = ReLU()(model)
            return(model)

        image = Input(shape=self.low_resolution_image_size)

        pre_residual = image
        if self.dimensionality == 2:
            pre_residual = Conv2D(filters=number_of_filters,
                                  kernel_size=9,
                                  strides=1,
                                  padding='same')(pre_residual)
        else:
            pre_residual = Conv3D(filters=number_of_filters,
                                  kernel_size=9,
                                  strides=1,
                                  padding='same')(pre_residual)

        residuals = build_residual_block(pre_residual,
          number_of_filters=self.number_of_filters_at_base_layer[0])
        for i in range(self.number_of_residual_blocks - 1):
            residuals = build_residual_block(residuals,
              number_of_filters=self.number_of_filters_at_base_layer[0])

        post_residual = residuals
        if self.dimensionality == 2:
            post_residual = Conv2D(filters=number_of_filters,
                                   kernel_size=3,
                                   strides=1,
                                   padding='same')(post_residual)
        else:
            post_residual = Conv3D(filters=number_of_filters,
                                  kernel_size=3,
                                  strides=1,
                                  padding='same')(post_residual)
        post_residual = BatchNormalization(momentum=0.8)(post_residual)
        model = Add()([post_residual, pre_residual])

        # upsampling

        if self.scale_factor >= 2:
            model = build_deconvolution_layer(model)
        if self.scale_factor >= 4:
            model = build_deconvolution_layer(model)
        if self.scale_factor == 8:
            model = build_deconvolution_layer(model)

        if self.dimensionality == 2:
            model = Conv2D(filters=self.number_of_channels,
                           kernel_size=9,
                           strides=1,
                           padding='same',
                           activation='tanh')(model)
        else:
            model = Conv3D(filters=self.number_of_channels,
                           kernel_size=9,
                           strides=1,
                           padding='same',
                           activation='tanh')(model)

        generator = Model(inputs=image, outputs=model)
        return(generator)

    def build_discriminator(self):

        def build_layer(input, number_of_filters, strides=1, kernel_size=3,
                        normalization=True):
            layer = input
            if self.dimensionality == 2:
                layer = Conv2D(filters=number_of_filters,
                               kernel_size=kernel_size,
                               strides=strides,
                               padding='same')(layer)
            else:
                layer = Conv2D(filters=number_of_filters,
                               kernel_size=kernel_size,
                               strides=strides,
                               padding='same')(layer)
            layer = LeakyReLU(alpha=0.2)(layer)
            if normalization == True:
                layer = BatchNormalization(momentum=0.8)(layer)
            return(layer)

        image = Input(shape=self.high_resolution_image_size)

        model = build_layer(image, self.number_of_filters_at_base_layer[1],
          normalization = False)
        model = build_layer(model, self.number_of_filters_at_base_layer[1],
          strides=2)
        model = build_layer(model, self.number_of_filters_at_base_layer[1] * 2)
        model = build_layer(model, self.number_of_filters_at_base_layer[1] * 2,
          strides=2)
        model = build_layer(model, self.number_of_filters_at_base_layer[1] * 4)
        model = build_layer(model, self.number_of_filters_at_base_layer[1] * 4,
          strides=2)
        model = build_layer(model, self.number_of_filters_at_base_layer[1] * 8)
        model = build_layer(model, self.number_of_filters_at_base_layer[1] * 8,
          strides=2)

        model = Dense(units=self.number_of_filters_at_base_layer[1] * 16)(model)
        model = LeakyReLU(alpha=0.2)(model)
        validity = Dense(units=1, activation = 'sigmoid')(model)

        discriminator = Model(inputs=image, outputs=validity)
        return(discriminator)

    def train(self, X_train_low_resolution, X_train_high_resolution, number_of_epochs, batch_size=128,
              sample_interval=None, sample_file_prefix='sample'):
        valid = np.ones((batch_size, *self.discriminator_patch_size))
        fake = np.zeros((batch_size, *self.discriminator_patch_size))

        for epoch in range(number_of_epochs):

            indices = np.random.randint(0, X_train_low_resolution.shape[0] - 1, batch_size)

            low_resolution_images = None
            high_resolution_images = None
            if self.dimensionality == 2:
                low_resolution_images = X_train_low_resolution[indices,:,:,:]
                high_resolution_images = X_train_high_resolution[indices,:,:,:]
            else:
                low_resolution_images = X_train_low_resolution[indices,:,:,:,:]
                high_resolution_images = X_train_high_resolution[indices,:,:,:,:]

            # train discriminator

            fake_high_resolution_images = self.generator.predict(low_resolution_images)

            d_loss_real = self.discriminator.train_on_batch(high_resolution_images, valid)
            d_loss_fake = self.discriminator.train_on_batch(fake_high_resolution_images, fake)
            d_loss = 0.5 * np.add(d_loss_real, d_loss_fake)

            # train generator
            g_loss = None
            if self.use_image_net_weights == True:
                image_features = self.vgg_model.predict(high_resolution_images)
                g_loss = self.combined_model.train_on_batch(
                  [low_resolution_images, high_resolution_images], [valid, image_features])
                print("Epoch ", epoch, ": [Discriminator loss: ", d_loss[0],
                    "] ", "[Generator loss: ", g_loss[0], "] ")
            else:
                g_loss = self.combined_model.train_on_batch(
                  [low_resolution_images, high_resolution_images], valid)
                print("Epoch ", epoch, ": [Discriminator loss: ", d_loss[0],
                    "] ", "[Generator loss: ", g_loss, "] ")


            if self.dimensionality == 2:
                if sample_interval != None:
                    if epoch % sample_interval == 0:

                        # Do a 2x3 grid
                        #
                        # low res image | high res image | original high res image
                        # low res image | high res image | original high res image

                        X = list()

                        index = np.random.randint(0, X_train_low_resolution.shape[0] - 1, 1)
                        low_resolution_image = X_train_low_resolution[index,:,:,:]
                        high_resolution_image = X_train_high_resolution[index,:,:,:]

                        X.append(self.generator.predict(low_resolution_image))
                        X.append(high_resolution_image)

                        index = np.random.randint(0, X_train_low_resolution.shape[0] - 1, 1)
                        low_resolution_image = X_train_low_resolution[index,:,:,:]
                        high_resolution_image = X_train_high_resolution[index,:,:,:]

                        X.append(self.generator.predict(low_resolution_image))
                        X.append(high_resolution_image)

                        plot_images = np.concatenate(X)
                        plot_images = 0.5 * plot_images + 0.5

                        titles = ['Predicted', 'Original']
                        figure, axes = plot.subplots(2, 2)

                        count = 0
                        for i in range(2):
                            for j in range(2):
                                axes[i, j].imshow(plot_images[count])
                                axes[i, j].set_title(titles[j])
                                axes[i, j].axis('off')
                                count += 1

                        image_file_name = sample_file_prefix + "_iteration" + str(epoch) + ".jpg"
                        dir_name = os.path.dirname(sample_file_prefix)
                        if not os.path.exists(dir_name):
                            os.mkdir(dir_name)
                        figure.savefig(image_file_name)
                        plot.close()




import ants
import numpy as np
import tensorflow.keras.backend as K

from . import extract_image_patches
from . import regression_match_image

from tensorflow.keras.models import load_model
import tensorflow as tf

import time
from os import path


def mse(x, y=None):
    """
    Mean square error of a single image or between two images.

    Arguments
    ---------
    x : input image
        ants input image

    y : input image
        ants input image

    Returns
    -------
    Value.

    Example
    -------
    >>> r16 = ants.image_read(ants.get_data("r16"))
    >>> r64 = ants.image_read(ants.get_data("r64"))
    >>> value = mse(r16, r64)
    """

    if y is None:
        x2 = x ** 2
        return x2.mean()
    else:
        diff2 = (x - y) ** 2
        return diff2.mean()


def mae(x, y=None):
    """
    Mean absolute error of a single image or between two images.

    Arguments
    ---------
    x : input image
        ants input image

    y : input image
        ants input image

    Returns
    -------
    Value

    Example
    -------
    >>> r16 = ants.image_read(ants.get_data("r16"))
    >>> r64 = ants.image_read(ants.get_data("r64"))
    >>> value = mae(r16, r64)
    """

    if y is None:
        xabs = x.abs()
        return xabs.mean()
    else:
        diffabs = (x-y).abs().mean()
        return diffabs.mean()

def psnr(x, y):
    """
    Peak signal-to-noise ratio between two images.

    Arguments
    ---------
    x : input image
        ants input image

    y : input image
        ants input image

    Returns
    -------
    Value

    Example
    -------
    >>> r16 = ants.image_read(ants.get_data("r16"))
    >>> r64 = ants.image_read(ants.get_data("r64"))
    >>> value = psnr(r16, r64)
    """

    value = 20 * np.log10(x.max()) - 10 * np.log10(mse(x, y))
    return value


def ssim(x, y, K=(0.01, 0.03)):
    """
    Structural similarity index (SSI) between two images.

    Implementation of the SSI quantity for two images proposed in

    Z. Wang, A.C. Bovik, H.R. Sheikh, E.P. Simoncelli. "Image quality
    assessment: from error visibility to structural similarity". IEEE TIP.
    13 (4): 600–612.

    Arguments
    ---------
    x : input image
        ants input image

    y : input image
        ants input image

    K : tuple of length 2
        tuple which contain SSI parameters meant to stabilize the formula
        in case of weak denominators.

    Returns
    -------
    Value

    Example
    -------
    >>> r16 = ants.image_read(ants.get_data("r16"))
    >>> r64 = ants.image_read(ants.get_data("r64"))
    >>> value = psnr(r16, r64)
    """

    global_max = np.max( ( x.max(), y.max()) )
    global_min = np.abs(min( ( x.min(), y.min())) )
    L = global_max - global_min

    C1 = (K[0] * L) ** 2
    C2 = (K[1] * L) ** 2
    C3 = C2 / 2

    mu_x = x.mean()
    mu_y = y.mean()

    mu_x_sq = mu_x * mu_x
    mu_y_sq = mu_y * mu_y
    mu_xy = mu_x * mu_y

    sigma_x_sq = (x * x).mean() - mu_x_sq
    sigma_y_sq = (y * y).mean() - mu_y_sq
    sigma_xy = (x * y).mean() - mu_xy

    numerator = (2 * mu_xy + C1) * (2 * sigma_xy + C2)
    denominator = (mu_x_sq + mu_y_sq + C1) * (sigma_x_sq + sigma_y_sq + C2)

    SSI = numerator / denominator

    return SSI


def gmsd(x, y):
    """
    Gradient magnitude similarity deviation

    A fast and simple metric that correlates to perceptual quality.

    Arguments
    ---------
    x : input image
        ants input image

    y : input image
        ants input image

    Returns
    -------
    Value

    Example
    -------
    >>> r16 = ants.image_read(ants.get_data("r16"))
    >>> r64 = ants.image_read(ants.get_data("r64"))
    >>> value = gmsd(r16, r64)
    """

    gx = ants.iMath(x, "Grad")
    gy = ants.iMath(y, "Grad")

    # see eqn 4 - 6 in https://arxiv.org/pdf/1308.3052.pdf

    constant = 0.0026
    gmsd_numerator = gx * gy * 2.0 + constant
    gmsd_denominator = gx ** 2 + gy ** 2 + constant
    gmsd = gmsd_numerator / gmsd_denominator

    product_dimension = 1
    for i in range(len(x.shape)):
        product_dimension *= x.shape[i]
    prefactor = 1.0 / product_dimension

    return np.sqrt(prefactor * ((gmsd - gmsd.mean()) ** 2).sum())


def apply_super_resolution_model_to_image(
    image,
    model,
    target_range=(-127.5, 127.5),
    batch_size=32,
    regression_order=None,
    verbose=False,
):

    """
    Apply a pretrained deep back projection model for super resolution.
    Helper function for applying a pretrained deep back projection model.
    Apply a patch-wise trained network to perform super-resolution. Can be applied
    to variable sized inputs. Warning: This function may be better used on CPU
    unless the GPU can accommodate the full image size. Warning 2: The global
    intensity range (min to max) of the output will match the input where the
    range is taken over all channels.

    Arguments
    ---------
    image : ANTs image
        input image.

    model : keras object or string
        pretrained keras model or filename.

    target_range : 2-element tuple
        a tuple or array defining the (min, max) of the input image
        (e.g., -127.5, 127.5).  Output images will be scaled back to original
        intensity. This range should match the mapping used in the training
        of the network.

    batch_size : integer
        Batch size used for the prediction call.

    regression_order : integer
        If specified, Apply the function regression_match_image with
        poly_order=regression_order.

    verbose : boolean
        If True, show status messages.

    Returns
    -------
    Super-resolution image upscaled to resolution specified by the network.

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> image_sr = apply_super_resolution_model_to_image(image, get_pretrained_network("dbpn4x"))
    """
    tflite_flag = False
    channel_axis = 0
    if K.image_data_format() == "channels_last":
        channel_axis = -1

    if target_range[0] > target_range[1]:
        target_range = target_range[::-1]

    start_time = time.time()
    if isinstance(model, str):
        if path.isfile(model):
            if verbose:
                print("Load model.")
            if path.splitext(model)[1] == '.tflite':
                interpreter = tf.lite.Interpreter(model)
                interpreter.allocate_tensors()
                input_details = interpreter.get_input_details()
                output_details = interpreter.get_output_details()
                shape_length = len(interpreter.get_input_details()[0]['shape'])
                tflite_flag = True
            else:    
                model = load_model(model)
                shape_length = len(model.input_shape)

            if verbose:
                elapsed_time = time.time() - start_time
                print("  (elapsed time: ", elapsed_time, ")")
        else:
            raise ValueError("Model not found.")
    else:
        shape_length = len(model.input_shape)

    
    if shape_length < 4 | shape_length > 5:
        raise ValueError("Unexpected input shape.")
    else:
        if shape_length == 5 & image.dimension != 3:
            raise ValueError("Expecting 3D input for this model.")
        elif shape_length == 4 & image.dimension != 2:
            raise ValueError("Expecting 2D input for this model.")

    if channel_axis == -1:
        channel_axis < shape_length
    if  tflite_flag:
        channel_size = interpreter.get_input_details()[0]['shape'][channel_axis]
    else:
        channel_size = model.input_shape[channel_axis]

    if channel_size != image.components:
        raise ValueError(
            "Channel size of model",
            str(channel_size),
            "does not match ncomponents=",
            str(image.components),
            "of the input image.",
        )

    image_patches = extract_image_patches(
        image,
        patch_size=image.shape,
        max_number_of_patches=1,
        stride_length=image.shape,
        return_as_array=True,
    )
    if image.components == 1:
        image_patches = np.expand_dims(image_patches, axis=-1)

    image_patches = image_patches - image_patches.min()
    image_patches = (
        image_patches / image_patches.max() * (target_range[1] - target_range[0])
        + target_range[0]
    )

    if verbose:
        print("Prediction")

    start_time = time.time()

    if  tflite_flag:
        image_patches = image_patches.astype('float32')
        interpreter.set_tensor(input_details[0]['index'], image_patches)
        interpreter.invoke()
        out = interpreter.tensor(output_details[0]['index'])
        prediction = out()
    else:
        prediction = model.predict(image_patches, batch_size=batch_size)

    if verbose:
        elapsed_time = time.time() - start_time
        print("  (elapsed time: ", elapsed_time, ")")

    if verbose:
        print("Reconstruct intensities")

    intensity_range = image.range()
    prediction = prediction - prediction.min()
    prediction = (
        prediction / prediction.max() * (intensity_range[1] - intensity_range[0])
        + intensity_range[0]
    )

    def slice_array_channel(input_array, slice, channel_axis=-1):
        if channel_axis == 0:
            if shape_length == 4:
                return input_array[slice, :, :, :]
            else:
                return input_array[slice, :, :, :, :]
        else:
            if shape_length == 4:
                return input_array[:, :, :, slice]
            else:
                return input_array[:, :, :, :, slice]

    expansion_factor = np.asarray(prediction.shape) / np.asarray(image_patches.shape)
    if channel_axis == 0:
        FIXME

    expansion_factor = expansion_factor[1 : (len(expansion_factor) - 1)]

    if verbose:
        print("ExpansionFactor:", str(expansion_factor))

    if image.components == 1:
        image_array = slice_array_channel(prediction, 0, channel_axis)
        prediction_image = ants.make_image(
            (np.asarray(image.shape) * np.asarray(expansion_factor)).astype(int),
            image_array,
        )
        if regression_order is not None:
            reference_image = ants.resample_image_to_target(image, prediction_image)
            prediction_image = regression_match_image(
                prediction_image, reference_image, poly_order=regression_order
            )
    else:
        image_component_list = list()
        for k in range(image.components):
            image_array = slice_array_channel(prediction, k, channel_axis)
            image_component_list.append(
                ants.make_image(
                    (np.asarray(image.shape) * np.asarray(expansion_factor)).astype(
                        int
                    ),
                    image_array,
                )
            )
        prediction_image = ants.merge_channels(image_component_list)

    prediction_image = ants.copy_image_info(image, prediction_image)
    ants.set_spacing(
        prediction_image,
        tuple(np.asarray(image.spacing) / np.asarray(expansion_factor)),
    )

    return prediction_image



import ants

import numpy as np
import random

def reconstruct_image_from_patches(patches,
                                   domain_image,
                                   stride_length=1,
                                   domain_image_is_mask=False):
    """
    Reconstruct image from a list of patches.

    Arguments
    ---------
    patches : list or array of patches
        List or array of patches defining an image.  Patches are assumed
        to have the same format as returned by extract_image_patches.

    domain_image : ANTs image
        Image or mask to define the geometric information of the
        reconstructed image.  If this is a mask image, the reconstruction will only
        use patches in the mask.

    stride_length:  integer or n-D tuple
        Defines the sequential patch overlap for max_number_of_patches = "all".
        Can be a image-dimensional vector or a scalar.

    domain_image_is_mask : boolean
        Boolean specifying whether the domain image is a
        mask used to limit the region of reconstruction from the patches.

    Returns
    -------
    An ANTs image.

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> image_patches = extract_image_patches(image, patch_size=(16, 16), stride_length=4)
    >>> reconstructed_image = reconstruct_image_from_patches(image_patches, image, stride_length=4)
    """

    image_size = domain_image.shape
    dimensionality = domain_image.dimension

    if dimensionality != 2 and dimensionality != 3:
        raise ValueError("Unsupported dimensionality.")

    is_list_patches = isinstance(patches, list)

    patch_size = ()
    number_of_image_components = 1
    if is_list_patches:
        patch_dimension = patches[0].shape
        patch_size = patch_dimension[0:dimensionality]
        if len(patch_dimension) > dimensionality:
            number_of_image_components = patch_dimension[dimensionality]
    else:
        patch_dimension = patches.shape
        patch_size = patch_dimension[1:( 2 + dimensionality - 1 )]
        if len(patch_dimension) > dimensionality + 1:
            number_of_image_components = patch_dimension[dimensionality + 1]

    mid_patch_index = tuple(np.int_(np.subtract(np.round(0.5 * (np.array(patch_size))), 1)))

    image_array = np.zeros((*image_size, number_of_image_components))

    stride_length_tuple = stride_length
    if isinstance(stride_length, int):
        stride_length_tuple = tuple(np.multiply(np.ones_like(patch_size), stride_length))
    elif len(stride_length) != dimensionality:
        raise ValueError("stride_length is not a scalar or vector of length dimensionality.")
    elif np.any(np.less(stride_length, 1)):
        raise ValueError("stride_length elements must be positive integers.")

    if domain_image_is_mask:
        mask_array = domain_image.numpy()
        mask_array[np.where(mask_array != 0)] = 1

    count = 0
    if dimensionality == 2:
        if np.all(np.equal(stride_length_tuple, 1)):
            for i in range(image_size[0] - patch_size[0] + 1):
                for j in range(image_size[1] - patch_size[1] + 1):
                    start_index = (i, j)
                    end_index = tuple(np.add(start_index, patch_size))

                    do_add = True
                    if domain_image_is_mask:
                        if mask_array[start_index[0]:end_index[0],
                                      start_index[1]:end_index[1]][mid_patch_index[0],
                                                                   mid_patch_index[1]] == 0:
                            do_add = False

                    if do_add:
                        patch = np.empty((1, 1))
                        if is_list_patches:
                            patch = patches[count]
                        else:
                            if number_of_image_components == 1:
                                patch = patches[count, :, :]
                            else:
                                patch = patches[count, :, :, :]

                        if number_of_image_components == 1:
                            patch = np.expand_dims(patch, axis = 2)

                        image_array[start_index[0]:end_index[0],
                                    start_index[1]:end_index[1], :][mid_patch_index[0],
                                                                    mid_patch_index[1], :] += patch
                        count += 1

            if not domain_image_is_mask:
                for i in range(image_size[0]):
                    for j in range(image_size[1]):
                        factor = (min(i + 1, patch_size[0], image_size[0] - i) *
                                  min(j + 1, patch_size[1], image_size[1] - j))

                        image_array[i, j, :] /= factor
        else:
            count_array = np.zeros(image_array.shape[0:dimensionality])
            for i in range(0, image_size[0] - patch_size[0] + 1, stride_length_tuple[0]):
                for j in range(0, image_size[1] - patch_size[1] + 1, stride_length_tuple[1]):
                    start_index = (i, j)
                    end_index = tuple(np.add(start_index, patch_size))

                    patch = np.empty((1, 1))
                    if is_list_patches:
                        patch = patches[count]
                    else:
                        if number_of_image_components == 1:
                            patch = patches[count, :, :]
                        else:
                            patch = patches[count, :, :, :]

                    if number_of_image_components == 1:
                        patch = np.expand_dims(patch, axis = 2)

                    image_array[start_index[0]:end_index[0],
                                start_index[1]:end_index[1], :] += patch
                    count_array[start_index[0]:end_index[0],
                                start_index[1]:end_index[1]] += np.ones(patch_size)
                    count += 1

            count_array[np.where(count_array == 0)] = 1
            for i in range(number_of_image_components):
                image_array[:, :, i] /= count_array

    else:

        if np.all(np.equal(stride_length_tuple, 1)):
            for i in range(image_size[0] - patch_size[0] + 1):
                for j in range(image_size[1] - patch_size[1] + 1):
                    for k in range(image_size[2] - patch_size[2] + 1):
                        start_index = (i, j, k)
                        end_index = tuple(np.add(start_index, patch_size))

                        do_add = True
                        if domain_image_is_mask:
                            if mask_array[start_index[0]:end_index[0],
                                          start_index[1]:end_index[1],
                                          start_index[2]:end_index[2]][mid_patch_index[0],
                                                                       mid_patch_index[1],
                                                                       mid_patch_index[2]] == 0:
                                do_add = False

                        if do_add:
                            patch = np.empty((1, 1, 1))
                            if is_list_patches:
                                patch = patches[count]
                            else:
                                if number_of_image_components == 1:
                                    patch = patches[count, :, :, :]
                                else:
                                    patch = patches[count, :, :, :, :]

                            if number_of_image_components == 1:
                                patch = np.expand_dims(patch, axis = 3)

                            image_array[start_index[0]:end_index[0],
                                        start_index[1]:end_index[1],
                                        start_index[2]:end_index[2], :][mid_patch_index[0],
                                                                        mid_patch_index[1],
                                                                        mid_patch_index[2], :] += patch
                            count += 1

            if not domain_image_is_mask:
                for i in range(image_size[0] + 1):
                    for j in range(image_size[1] + 1):
                        for k in range(image_size[2] + 1):
                            factor = (min(i + 1, patch_size[0], image_size[0] - i) *
                                      min(j + 1, patch_size[1], image_size[1] - j) *
                                      min(k + 1, patch_size[2], image_size[2] - k))

                            image_array[i, j, k, :] /= factor
        else:
            count_array = np.zeros(image_array.shape[0:dimensionality])
            for i in range(0, image_size[0] - patch_size[0] + 1, stride_length_tuple[0]):
                for j in range(0, image_size[1] - patch_size[1] + 1, stride_length_tuple[1]):
                    for k in range(0, image_size[2] - patch_size[2] + 1, stride_length_tuple[2]):
                        start_index = (i, j, k)
                        end_index = tuple(np.add(start_index, patch_size))

                        patch = np.empty((1, 1, 1))
                        if is_list_patches:
                            patch = patches[count]
                        else:
                            if number_of_image_components == 1:
                                patch = patches[count, :, :, :]
                            else:
                                patch = patches[count, :, :, :, :]

                        if number_of_image_components == 1:
                            patch = np.expand_dims(patch, axis = 3)

                        image_array[start_index[0]:end_index[0],
                                    start_index[1]:end_index[1],
                                    start_index[2]:end_index[2], :] += patch
                        count_array[start_index[0]:end_index[0],
                                    start_index[1]:end_index[1],
                                    start_index[2]:end_index[2]] += np.ones(patch_size)
                        count += 1

            count_array[np.where(count_array == 0)] = 1
            for i in range(number_of_image_components):
                image_array[:, :, :, i] /= count_array

    if dimensionality == 2:
        image_array = np.transpose(image_array, [2, 0, 1])
    else:
        image_array = np.transpose(image_array, [3, 0, 1, 2])

    image_array = np.squeeze(image_array)

    reconstructed_image = ants.from_numpy(image_array,
                                          origin=domain_image.origin,
                                          spacing=domain_image.spacing,
                                          direction=domain_image.direction,
                                          has_components=(number_of_image_components != 1))

    return(reconstructed_image)








import ants

import numpy as np
import random

def extract_image_patches(image,
                          patch_size,
                          max_number_of_patches='all',
                          stride_length=1,
                          mask_image=None,
                          random_seed=None,
                          return_as_array=False,
                          randomize=True):
    """
    Extract 2-D or 3-D image patches.

    Arguments
    ---------
    image : ANTsImage
        Input image with one or more components.

    patch_size: n-D tuple (depending on dimensionality).
        Width, height, and depth (if 3-D) of patches.

    max_number_of_patches:  integer or string
        Maximum number of patches returned.  If "all" is specified, then
        all patches in sequence (defined by the stride_length are extracted.

    stride_length:  integer or n-D tuple
        Defines the sequential patch overlap for max_number_of_patches = "all".
        Can be a image-dimensional vector or a scalar.

    mask_image:  ANTsImage (optional)
        Optional image specifying the sampling region for
        the patches when max_number_of_patches does not equal "all".
        The way we constrain patch selection using a mask is by forcing
        each returned patch to have a masked voxel at its center.

    random_seed: integer (optional)
        Seed value that allows reproducible patch extraction across runs.

    return_as_array: boolean
        Specifies the return type of the function.  If
        False (default) the return type is a list where each element is
        a single patch.  Otherwise the return type is an array of size
        dim( number_of_patches, patch_size ).

    randomize: boolean
        Boolean controlling whether we randomize indices when masking.

    Returns
    -------
    A list (or array) of patches.

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data('r16'))
    >>> image_patches = extract_image_patches(image, patch_size=(32, 32))
    """

    if random_seed is not None:
        random.seed(random_seed)

    image_size = image.shape
    dimensionality = image.dimension

    if dimensionality != 2 and dimensionality != 3:
        raise ValueError("Unsupported dimensionality.")

    number_of_image_components = image.components

    if len(image_size) != len(patch_size):
        raise ValueError("Mismatch between the image size and the specified patch size.")

    if patch_size > image_size:
        raise ValueError("Patch size is greater than the image size.")

    image_array = image.numpy()

    patch_list = []
    patch_array = np.empty([1, 1])
    mid_patch_index = tuple(np.int_(np.subtract(np.round(0.5 * (np.array(patch_size))), 1)))

    number_of_extracted_patches = max_number_of_patches

    if isinstance(max_number_of_patches, str) and max_number_of_patches.lower() == 'all':
        stride_length_tuple = stride_length
        if isinstance(stride_length, int):
            stride_length_tuple = tuple(np.multiply(np.ones_like(patch_size), stride_length))
        elif len(stride_length) != dimensionality:
            raise ValueError("stride_length is not a scalar or vector of length dimensionality.")
        elif np.any(np.less(stride_length, 1)):
            raise ValueError("stride_length elements must be positive integers.")

        number_of_extracted_patches = 1

        indices = []
        for d in range(dimensionality):
            indices.append(range(0, image_size[d] - patch_size[d] + 1, stride_length_tuple[d]))
            number_of_extracted_patches *= len(indices[d])

        if return_as_array:
            if number_of_image_components == 1:
                patch_array = np.zeros((number_of_extracted_patches, *patch_size))
            else:
                patch_array = np.zeros((number_of_extracted_patches, *patch_size, number_of_image_components))

        count = 0
        if dimensionality == 2:
            for i in indices[0]:
                for j in indices[1]:
                    start_index = (i, j)
                    end_index = tuple(np.add(start_index, patch_size))

                    if number_of_image_components == 1:
                        patch = image_array[start_index[0]:end_index[0],
                                            start_index[1]:end_index[1]]
                    else:
                        patch = image_array[start_index[0]:end_index[0],
                                            start_index[1]:end_index[1],:]

                    if return_as_array:
                        if number_of_image_components == 1:
                            patch_array[count, :, :] = patch
                        else:
                            patch_array[count, :, :, :] = patch
                    else:
                        patch_list.append(patch)

                    count += 1
        else:
            for i in indices[0]:
                for j in indices[1]:
                    for k in indices[2]:
                        start_index = (i, j, k)
                        end_index = tuple(np.add(start_index, patch_size))

                        if number_of_image_components == 1:
                            patch = image_array[start_index[0]:end_index[0],
                                                start_index[1]:end_index[1],
                                                start_index[2]:end_index[2]]
                        else:
                            patch = image_array[start_index[0]:end_index[0],
                                                start_index[1]:end_index[1],
                                                start_index[2]:end_index[2],:]

                        if return_as_array:
                            if number_of_image_components == 1:
                                patch_array[count, :, :, :] = patch
                            else:
                                patch_array[count, :, :, :, :] = patch
                        else:
                            patch_list.append(patch)

                        count += 1
    else:

        random_indices = np.zeros((max_number_of_patches, dimensionality), dtype=int)

        if mask_image != None:
            mask_array = mask_image.numpy()
            mask_array[np.where(mask_array != 0)] = 1

            # The way we constrain patch selection using a mask is by assuming that
            # each patch must have a masked voxel at its midPatchIndex.

            mask_indices = np.column_stack(np.where(mask_array != 0))

            for d in range(dimensionality):
                shifted_mask_indices = np.subtract(mask_indices[:, d], mid_patch_index[d])
                outside_indices = np.where(shifted_mask_indices < 0)
                mask_indices = np.delete(mask_indices, outside_indices, axis = 0)
                shifted_mask_indices = np.add(mask_indices[d], mid_patch_index[d])
                outside_indices = np.where(shifted_mask_indices >= image_size[d])
                mask_indices = np.delete(mask_indices, outside_indices, axis = 0)

            # After pruning the mask indices, which were originally defined in terms of the
            # mid_patch_index, we subtract the midPatchIndex so that it's now defined at the
            # corner for patch selection.

            for d in range(dimensionality):
                mask_indices[:, d] = np.subtract(mask_indices[:, d], mid_patch_index[d])

            number_of_extracted_patches = min(max_number_of_patches, mask_indices.shape[0])

            if randomize:
                random_indices = mask_indices[
                  random.sample(range(mask_indices.shape[0] + 1), number_of_extracted_patches), :]
            else:
                random_indices = mask_indices

        else:

            for d in range(dimensionality):
                random_indices[:, d] = random.sample(range(image_size[d] - patch_size[d] + 1), max_number_of_patches)

        if return_as_array:
            if number_of_image_components == 1:
                patch_array = np.zeros((number_of_extracted_patches, *patch_size))
            else:
                patch_array = np.zeros((number_of_extracted_patches, *patch_size, number_of_image_components))

        start_index = np.ones( dimensionality )
        for i in range(number_of_extracted_patches):
            start_index = random_indices[i, :]
            end_index = np.add(start_index, patch_size)

            if dimensionality == 2:
                if number_of_image_components == 1:
                    patch = image_array[start_index[0]:end_index[0],
                                        start_index[1]:end_index[1]]
                else:
                    patch = image_array[start_index[0]:end_index[0],
                                        start_index[1]:end_index[1], :]
            else:
                if number_of_image_components == 1:
                    patch = image_array[start_index[0]:end_index[0],
                                        start_index[1]:end_index[1],
                                        start_index[2]:end_index[2]]
                else:
                    patch = image_array[start_index[0]:end_index[0],
                                        start_index[1]:end_index[1],
                                        start_index[2]:end_index[2], :]

            if return_as_array:
                if dimensionality == 2:
                    if number_of_image_components == 1:
                        patch_array[i, :, :] = patch
                    else:
                        patch_array[i, :, :, :] = patch
                else:
                    if number_of_image_components == 1:
                        patch_array[i, :, :] = patch
                    else:
                        patch_array[i, :, :, :, :] = patch
            else:
                patch_list.append(patch)

    if return_as_array:
       return(patch_array)
    else:
       return(patch_list)




import tensorflow as tf

import tensorflow.keras.backend as K

from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import (Input, Concatenate, Dense, Activation,
                                     BatchNormalization, Reshape, Dropout,
                                     Flatten, LeakyReLU, Conv2D, Conv3D)
from tensorflow.keras.optimizers import Adam

from . import (create_convolutional_autoencoder_model_2d,
               create_convolutional_autoencoder_model_3d)

from ..utilities import ResampleTensorLayer2D, ResampleTensorLayer3D

import numpy as np
import os

import ants

class DeepConvolutionalGanModel(object):
    """
    GAN model using CNNs

    Deep convolutional generative adverserial network from the paper:

      https://arxiv.org/abs/1511.06434

    and ported from the Keras (python) implementation:

      https://github.com/eriklindernoren/Keras-GAN/blob/master/dcgan/dcgan.py

    Arguments
    ---------
    input_image_size : tuple
        Used for specifying the input tensor shape.  The shape (or dimension) of
        that tensor is the image dimensions followed by the number of channels
        (e.g., red, green, and blue).

    latent_dimension : integer

    Returns
    -------
    Keras model
        A Keras model defining the network.
    """

    def __init__(self, input_image_size, latent_dimension=100):
        super(DeepConvolutionalGanModel, self).__init__()

        self.input_image_size = input_image_size
        self.latent_dimension = latent_dimension

        self.dimensionality = None
        if len(self.input_image_size) == 3:
            self.dimensionality = 2
        elif len(self.input_image_size) == 4:
            self.dimensionality = 3
        else:
            raise ValueError("Incorrect size for input_image_size.")

        optimizer = Adam(lr=0.0002, beta_1=0.5)

        self.discriminator = self.build_discriminator()

        self.discriminator.compile(loss='binary_crossentropy',
                                   optimizer=optimizer, metrics=['acc'])
        self.discriminator.trainable = False

        self.generator = self.build_generator()

        z = Input(shape=(self.latent_dimension,))
        image = self.generator(z)

        validity = self.discriminator(image)

        self.combined_model = Model(inputs=z, outputs=validity)
        self.combined_model.compile(loss='binary_crossentropy',
                                    optimizer=optimizer)

    def build_generator(self, number_of_filters_per_layer=(128, 64), kernel_size=3):

        model = Sequential()

        # To build the generator, we create the reverse encoder model
        # and simply build the reverse model

        encoder = None
        if self.dimensionality == 2:
             autoencoder, encoder = create_convolutional_autoencoder_model_2d(
                          input_image_size=self.input_image_size,
                          number_of_filters_per_layer=(*(number_of_filters_per_layer[::-1]), self.latent_dimension),
                          convolution_kernel_size=(5, 5),
                          deconvolution_kernel_size=(5, 5))
        else:
             autoencoder, encoder = create_convolutional_autoencoder_model_3d(
                          input_image_size=self.input_image_size,
                          number_of_filters_per_layer=(*(number_of_filters_per_layer[::-1]), self.latent_dimension),
                          convolution_kernel_size=(5, 5, 5),
                          deconvolution_kernel_size=(5, 5, 5))

        encoder_layers = encoder.layers

        penultimate_layer = encoder_layers[len(encoder_layers) - 2]

        model.add(Dense(units=penultimate_layer.output_shape[1],
                        input_dim=self.latent_dimension,
                        activation="relu"))

        conv_layer = encoder_layers[len(encoder_layers) - 3]
        resampled_size = conv_layer.output_shape[1:(self.dimensionality + 2)]
        model.add(Reshape(resampled_size))

        count = 0
        for i in range(len(encoder_layers) - 3, 1, -1):
            conv_layer = encoder_layers[i]
            resampled_size = conv_layer.output_shape[1:(self.dimensionality + 1)]

            if self.dimensionality == 2:
                model.add(ResampleTensorLayer2D(shape=resampled_size,
                                                interpolation_type='linear'))
                model.add(Conv2D(filters=number_of_filters_per_layer[count],
                                 kernel_size=kernel_size,
                                 padding='same'))
            else:
                model.add(ResampleTensorLayer3D(shape=resampled_size,
                                                interpolation_type='linear'))
                model.add(Conv3D(filters=number_of_filters_per_layer[count],
                                 kernel_size=kernel_size,
                                 padding='same'))
            model.add(BatchNormalization(momentum=0.8))
            model.add(Activation(activation='relu'))
            count += 1

        number_of_channels = self.input_image_size[-1]
        spatial_dimensions = self.input_image_size[:self.dimensionality]

        if self.dimensionality == 2:
            model.add(ResampleTensorLayer2D(shape=spatial_dimensions,
                                            interpolation_type='linear'))
            model.add(Conv2D(filters=number_of_channels,
                             kernel_size=kernel_size,
                             padding='same'))
        else:
            model.add(ResampleTensorLayer3D(shape=spatial_dimensions,
                                            interpolation_type='linear'))
            model.add(Conv3D(filters=number_of_channels,
                             kernel_size=kernel_size,
                             padding='same'))

        model.add(Activation(activation="tanh"))

        noise = Input(shape=(self.latent_dimension,))
        image = model(noise)

        generator = Model(inputs=noise, outputs=image)
        return(generator)

    def build_discriminator(self, number_of_filters_per_layer=(32, 64, 128, 256),
                            kernel_size=3, dropout_rate=0.25):
        model = Sequential()

        for i in range(len(number_of_filters_per_layer)):
            if self.dimensionality == 2:
                model.add(Conv2D(input_shape=self.input_image_size,
                                 filters=number_of_filters_per_layer[i],
                                 kernel_size = kernel_size,
                                 strides = 2,
                                 padding='same'))
            else:
                model.add(Conv3D(input_shape=self.input_image_size,
                                 filters=number_of_filters_per_layer[i],
                                 kernel_size = kernel_size,
                                 strides = 2,
                                 padding='same'))

            model.add(BatchNormalization(momentum=0.8))
            model.add(LeakyReLU(alpha=0.2))
            model.add(Dropout(rate=dropout_rate))

        model.add(Flatten())
        model.add(Dense(units=1, activation='sigmoid'))

        image = Input(shape=self.input_image_size)

        validity = model(image)

        discriminator = Model(inputs=image, outputs=validity)

        return(discriminator)

    def train(self, X_train, number_of_epochs, batch_size=128,
              sample_interval=None, sample_file_prefix='sample'):
        valid = np.ones((batch_size, 1))
        fake = np.zeros((batch_size, 1))

        for epoch in range(number_of_epochs):

            # train discriminator

            indices = np.random.randint(0, X_train.shape[0] - 1, batch_size)
            X_valid_batch = X_train[indices]

            noise = np.random.normal(0, 1, (batch_size, self.latent_dimension))
            X_fake_batch = self.generator.predict(noise)

            d_loss_real = self.discriminator.train_on_batch(X_valid_batch, valid)
            d_loss_fake = self.discriminator.train_on_batch(X_fake_batch, fake)
            d_loss = 0.5 * np.add(d_loss_real, d_loss_fake)

            # train generator

            noise = np.random.normal(0, 1, (batch_size, self.latent_dimension))
            g_loss = self.combined_model.train_on_batch(noise, valid)

            print("Epoch ", epoch, ": [Discriminator loss: ", d_loss[0],
                  " acc: ", d_loss[1], "] ", "[Generator loss: ", g_loss)

            if self.dimensionality == 2:
                if sample_interval != None:
                    if epoch % sample_interval == 0:

                        # Do a 5x5 grid

                        predicted_batch_size = 5 * 5
                        noise = np.random.normal(0, 1, (predicted_batch_size, self.latent_dimension))
                        X_generated = self.generator.predict(noise)

                        # Convert to [0,255] to write as jpg using ANTsPy

                        X_generated = (255 * (X_generated - X_generated.min()) /
                          (X_generated.max() - X_generated.min()))
                        X_generated = np.squeeze(X_generated)
                        X_generated = np.uint8(X_generated)

                        X_tiled = np.zeros((5 * X_generated.shape[1], 5 * X_generated.shape[2]), dtype=np.uint8)
                        for i in range(5):
                            indices_i = (i * X_generated.shape[1], (i + 1) * X_generated.shape[1])
                            for j in range(5):
                                indices_j = (j * X_generated.shape[2], (j + 1) * X_generated.shape[2])
                                X_tiled[indices_i[0]:indices_i[1], indices_j[0]:indices_j[1]] = \
                                  np.squeeze(X_generated[i * 5 + j, :, :])

                        X_generated_image = ants.from_numpy(np.transpose(X_tiled))

                        image_file_name = sample_file_prefix + "_iteration" + str(epoch) + ".jpg"
                        dir_name = os.path.dirname(sample_file_prefix)
                        if not os.path.exists(dir_name):
                            os.mkdir(dir_name)

                        print("   --> writing sample image: ", image_file_name)
                        ants.image_write(X_generated_image, image_file_name)







import ants

import numpy as np

def simulate_bias_field(domain_image,
                        number_of_points=10,
                        sd_bias_field=1.0,
                        number_of_fitting_levels=4,
                        mesh_size=1):
    """
    Simulate random bias field

    Low frequency, spatial varying simulated random bias field using
    random points and B-spline fitting.

    Arguments
    ---------
    domain_image : ANTsImage
        Image to define the spatial domain of the bias field.

    number_of_points : integer
        Number of randomly defined points to define the bias field
        (default = 10).

    sd_bias_field : float
        Characterize the standard deviation of the amplitude (default = 1).

    number_of_fitting_levels : integer
        B-spline fitting parameter.

    clamp_end_points : integer or tuple
        B-spline fitting parameter.

    Returns
    -------
    ANTs image

    Example
    -------
    >>> import ants
    >>> image = ants.image_read(ants.get_ants_data("r64"))
    >>> bias_field_image = simulate_bias_field( image )
    """

    dimension = domain_image.dimension
    origin = ants.get_origin(domain_image)
    spacing = ants.get_spacing(domain_image)
    direction = ants.get_direction(domain_image)
    shape = domain_image.shape

    min_spatial_domain = origin
    max_spatial_domain = origin + (np.array(shape) - 1.0) * spacing

    scattered_data = np.zeros((number_of_points,1))
    parametric_data = np.zeros((number_of_points, dimension))

    scattered_data[:,0] = np.random.normal(loc=0.0, scale=sd_bias_field, size=number_of_points)
    for d in range(dimension):
        parametric_data[:,d] = np.random.uniform(low=min_spatial_domain[d],
            high=max_spatial_domain[d], size=number_of_points)

    if isinstance(mesh_size, int):
        mesh_size = np.tile(mesh_size, dimension)

    bias_field = ants.fit_bspline_object_to_scattered_data(scattered_data,
        parametric_data, parametric_domain_origin=origin, parametric_domain_spacing=spacing,
        parametric_domain_size=shape, number_of_fitting_levels=number_of_fitting_levels,
        mesh_size=mesh_size)

    ants.set_direction(bias_field, direction)

    return bias_field